from typing import Any
from dataclasses import dataclass
from .....utils import DictUtils, Crypto, get_object_model_class, pydash, init_class_kwargs
from ....base.firebase_collection import (FirebaseCollectionBase, getTimestamp)
from .code_fields import CodeFields, CodeFieldsProps, STANDARD_FIELDS, CODE_COLLECTION, get_code_expire_at, is_expired
from .....utils.response import ApiResponse, Error, ResponseStatus, StatusCode

@dataclass(init=False)
class Code(FirebaseCollectionBase):
    id: str
    code: str
    target: str
    code_type: str
    createdAt: any
    expired_at: str
    expires_in: int

    def __init__(self, code=None, **kwargs):
        if type(code) is str or type(code) is int:
            code = {"code": str(code), "code_type": "email_validation"} 
        (cls_object, keys, data_args) = init_class_kwargs(self, code, STANDARD_FIELDS, CodeFieldsProps, CODE_COLLECTION, ['id'], **kwargs)
        super().__init__(**data_args);
        for key in keys:
            setattr(self, key, cls_object[key]) 


    @staticmethod
    def generate_model(_key_="default_value"):
        code = {};
        props = CodeFieldsProps
        for k in DictUtils.get_keys(props):
            code[k] = props[k][_key_];
        return code;

    @staticmethod
    def from_dict(code: Any=None, db=None, collection_name=None) -> 'Code':
        cls_object, keys = get_object_model_class(code, Code, CodeFieldsProps);
        _client = Code(cls_object, db=db, collection_name=collection_name)
        return _client

    def query(self, query_params: list, first=False, limit=None):
        result = [];
        query_result = self.custom_query(query_params, first=first, limit=limit)
        if bool(query_result):
            if first is True:
                return Code.from_dict(code=query_result, db=self.db, collection_name=self.collection_name)
            else:
                for doc in query_result:
                    result.append(Code.from_dict(code=doc, db=self.db, collection_name=self.collection_name))
                return result
        if first:
                return None
        return [];

    def getCode(self) -> 'Code':
        if bool(self.id):
            doc = self.document_reference().get();
            if doc.exists is False:
                return None;
            return Code.from_dict(doc.to_dict(), db=self.db, collection_name=self.collection_name);
        if bool(self.code) and bool(self.code_type):
            params = [{"key": CodeFields.code, "comp": "==", "value": self.code}, {"key": CodeFields.code_type, "comp": "==", "value": self.code_type}]
            return self.query(params, True)
        return None;

    @staticmethod
    def generate(**kwargs) -> 'ApiResponse':
        data_dict = DictUtils.pick_fields(kwargs, CodeFields.filtered_keys('pickable', True));
        code_model = Code.from_dict(DictUtils.merge_dict(data_dict, Code.generate_model()));
        
        if bool(code_model.to_json()) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Empty request","INVALID_REQUEST", 1)).to_json()
        
        if bool(code_model.code) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("name required","INVALID_REQUEST", 1));

        code_model.id = Crypto().generate_id(code_model.target+"~"+code_model.code_type+"~");
        code_model.expired_at = get_code_expire_at(code_model.expires_in).isoformat()
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, code_model, None);
    @staticmethod
    def get_id(cls, target, type):
        return Crypto().generate_id(target+"~"+type+"~");
    @staticmethod
    def create(**kwargs):
        # print('creating code', kwargs)
        authorization_code = Code().generate(**kwargs);
        # print('authorization_code', authorization_code)
        if authorization_code.status == ResponseStatus.ERROR:
            return authorization_code
        
        authorization_code_model: Code = authorization_code.data;
        docRef = Code(authorization_code_model.to_json()).document_reference();
        # if docRef.get().exists:
        #     return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Code with name {authorization_code.id} already exist","INVALID_REQUEST", 1));
        
        docRef.set({**authorization_code_model.to_json(), "createdAt": getTimestamp()}, merge=True);
        created_campaign = authorization_code_model.getCode()
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, created_campaign.to_json(), None);
    
    def update(self, data):
        try:
            last_value = self.to_json();
            filtered_value = pydash.pick(data, CodeFields.filtered_keys('editable', True));
            new_value = DictUtils.merge_dict(filtered_value, self.to_json());
            changed_fields = DictUtils.get_changed_field(last_value, new_value);
            data_update = DictUtils.dict_from_keys(filtered_value, changed_fields);
            if bool(data_update) is False:
                return None;
            self.document_reference().set(data_update, merge=True)
            return DictUtils.dict_from_keys(self.getCode().to_json(), changed_fields);
        except Exception as e:
            print('Exception while update authorization code', e)
            return None;
    
    def use_code(self):
        code = self.getCode();
        if code is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Invalid code","INVALID_REQUEST", 1));
        if code.is_expired():
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Expired code","INVALID_REQUEST", 1));
        remove = code.remove()
        if remove.status == ResponseStatus.ERROR:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"An error occurated","INVALID_REQUEST", 1));
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, {"message": "success"}, None);

    def remove(self):
        try:
            if self.id is None:
                return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Cannot identify Authorization code with id {self.id} to delete","INVALID_REQUEST", 1));
            deleted = self.document_reference().delete();
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, {"message": f"Code {self.id} deleted"}, None);
        except Exception as e:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"An error occurated while removing authorization code with id {self.id} {str(e)}","INVALID_REQUEST", 1));

    def is_expired(self):
        return is_expired(self.expired_at)


from dataclasses import dataclass
from typing import Any
from  ....utils.utils import DictUtils, BaseFieldsClass, BaseClass,init_class_kwargs,get_object_model_class
@dataclass
class VideoAssetFields(BaseFieldsClass):
    id = "id"
    assetId = "assetId"
    videoId = "videoId"
    name = "name"
    thumbnails = "thumbnails"
    previewUrl = "previewUrl"
    snippet = "snippet"
    owner = "owner"
    accountId = "accountId"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

@dataclass(init=False)
class VideoAsset(BaseClass):
    id: str
    name: str
    assetId: int
    videoId: str
    name: str
    thumbnails: str
    previewUrl: str
    snippet: str
    owner: str
    accountId: str

    def __init__(self, asset=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, asset, None, None, None, [], **kwargs)
        for key in keys:
            setattr(self, key, cls_object[key]) 
    def to_dict(self, field=None):
        return DictUtils.remove_none_values({
            "id": self.id,
            "name": self.name,
            "assetId": self.assetId,
            "videoId": self.videoId,
            "thumbnails": self.thumbnails,
            "previewUrl": self.previewUrl,
            "snippet": self.snippet,
            "owner": self.owner,
            "accountId": self.accountId
        })
    @staticmethod
    def from_dict(obj: Any) -> 'VideoAsset':
        cls_object, keys = get_object_model_class(obj, VideoAsset, None);
        return VideoAsset(cls_object)




from ...base.firebase_collection import FirebaseCollectionBase, getTimestamp
from ....utils import (DictUtils, get_object_model_class, init_class_kwargs, Crypto, get_request_fields, BaseClass)
from ....utils.response import ApiResponse, Error, ResponseStatus, StatusCode
from .user_fields import UserFields, UserFieldProps, PhoneInfosFields, PhoneInfosFieldProps, PHONE_FIELDS, USER_FIELDS, USERS_COLLECTION
from typing import List
from typing import Any
from dataclasses import dataclass
from firebase_admin import auth as admin_auth
from adafri.v1.auth.firebase_auth import create_firebase_user
from firebase_admin.exceptions import FirebaseError
import pydash


def filter_user_request_fields(fields, default_fields = USER_FIELDS):
    request_fields = get_request_fields(fields, default_fields, [default_fields[0]])
    if request_fields is None or bool(request_fields) is False:
        request_fields = [default_fields[0]]
    return request_fields
@dataclass
class Account:
    canManageClients: bool
    childs: List[object]
    currenyCode: str
    customerId: int
    dateTimeZone: str
    name: str
    testAccount: bool

    @staticmethod
    def from_dict(obj: Any) -> 'Account':
        _canManageClients = DictUtils.pick(obj, "canManageClients", bool)
        _childs = [y for y in DictUtils.pick(obj, "childs", list)]
        _currenyCode = str(DictUtils.pick(obj, "currenyCode", str))
        _customerId = int(DictUtils.pick(obj, "customerId", int))
        _dateTimeZone = str(DictUtils.pick(obj, "dateTimeZone",str))
        _name = str(DictUtils.pick(obj, "name", str))
        _testAccount = bool(DictUtils.pick(obj, "testAccount", bool))
        return Account(_canManageClients, _childs, _currenyCode, _customerId, _dateTimeZone, _name, _testAccount)

@dataclass
class Country:
    areaCodes: List[object]
    dialCode: str
    flagClass: str
    htmlId: str
    iso2: str
    name: str
    placeHolder: str
    priority: int

    @staticmethod
    def from_dict(obj: Any) -> 'Country':
        _areaCodes = [y for y in DictUtils.pick(obj, "areaCodes", list)]
        _dialCode = str(DictUtils.pick(obj, "dialCode", str))
        _flagClass = str(DictUtils.pick(obj, "flagClass",str))
        _htmlId = str(DictUtils.pick(obj, "htmlId", str))
        _iso2 = str(DictUtils.pick(obj, "iso2", str))
        _name = str(DictUtils.pick(obj, "name", str))
        _placeHolder = str(DictUtils.pick(obj, "placeHolder", str))
        _priority = int(DictUtils.pick(obj, "priority", int))
        return Country(_areaCodes, _dialCode, _flagClass, _htmlId, _iso2, _name, _placeHolder, _priority)

@dataclass
class Credential:
    refresh_token: str
    scopes: List[str]
    token: str
    token_uri: str

    @staticmethod
    def from_dict(obj: Any) -> 'Credential':
        _refresh_token = str(DictUtils.pick(obj, "refresh_token", str))
        _scopes = [y for y in DictUtils.pick(obj, "scopes", list)]
        _token = str(DictUtils.pick(obj, "token", str))
        _token_uri = str(DictUtils.pick(obj, "token_uri", str))
        return Credential(_refresh_token, _scopes, _token, _token_uri)

@dataclass
class DeviceInfo:
    browser: str
    browser_version: str
    device: str
    os: str
    os_version: str
    userAgent: str

    @staticmethod
    def from_dict(obj: Any) -> 'DeviceInfo':
        _browser = str(DictUtils.pick(obj, "browser", str))
        _browser_version = str(DictUtils.pick(obj, "browser_version", str))
        _device = str(DictUtils.pick(obj, "device", str))
        _os = str(DictUtils.pick(obj, "os", str))
        _os_version = str(DictUtils.pick(obj, "os_version", str))
        _userAgent = str(DictUtils.pick(obj, "userAgent", str))
        return DeviceInfo(_browser, _browser_version, _device, _os, _os_version, _userAgent)

@dataclass
class PartenerData:
    id: str
    text: str

    @staticmethod
    def from_dict(obj: Any) -> 'PartenerData':
        _id = str(DictUtils.pick(obj, "id", str))
        _text = str(DictUtils.pick(obj, "text", str))
        return PartenerData(_id, _text)

@dataclass(init=False)
class PhoneInfo(BaseClass):
    countryCode: str
    dialCode: str
    e164Number: str
    internationalNumber: str
    nationalNumber: str
    number: str
    __baseFields: PhoneInfosFieldProps

    def __init__(self, phoneInfo=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, phoneInfo, PHONE_FIELDS, PhoneInfosFieldProps, None, [], **kwargs)
        super().__init__(**data_args);
        for key in keys:
            setattr(self, key, cls_object[key]) 


    @staticmethod
    def from_dict(obj: Any) -> 'PhoneInfo':
        cls_object, keys = get_object_model_class(obj, PhoneInfo, PhoneInfosFieldProps);
        _phone = PhoneInfo(cls_object)
        return _phone
    
    def to_dict(self, field=None):
        # print('to dict', self)
        return DictUtils.remove_none_values({
            "countryCode": self.countryCode,
            "dialCode": self.dialCode,
            "e164Number": self.e164Number,
            "internationalNumber": self.internationalNumber,
            "nationalNumber": self.nationalNumber,
            "number": self.number
        })

@dataclass
class PlateformRole:
    id: str
    partenerData: PartenerData
    text: str

    @staticmethod
    def from_dict(obj: Any) -> 'PlateformRole':
        _id = str(DictUtils.pick(obj, "id", str))
        _partenerData = PartenerData.from_dict(DictUtils.pick(obj, "partenerData", dict))
        _text = str(DictUtils.pick(obj, "text", str))
        return PlateformRole(_id, _partenerData, _text)


@dataclass(init=False)
class User(FirebaseCollectionBase):
    account_value: int = None;
    addresse: str
    auth_code: str
    authorizedPush: bool
    country: Country
    deviceInfo: DeviceInfo
    displayName: str
    email: str
    entrepriseName: str
    entrepriseUrl: str
    first_name: str
    hasApprouvedPolicy: bool
    isConnectWithMailAndPassword: bool
    isCorporate: bool
    isDesktopDevice: bool
    isMobile: bool
    isParticular: bool
    isTablet: bool
    last_name: str
    phoneInfo: PhoneInfo
    photoURL: str
    plateformRole: PlateformRole
    postal: str
    profileCompleted: bool
    showPushToken: bool
    telephone: str
    token: List[str]
    uid: str
    user_type: str
    password: str
    provider: str
    status: str
    _emailValidationSendDate: str
    _pwResetSendDate: str
    businessType: str
    phoneVerified: bool
    emailVerified: bool
    postalCode: str
    __baseFields: UserFieldProps

    def __init__(self, user=None, **kwargs):
        value = user;
        if type(user) is str:
            value = {"uid": user, "id": user}
        (cls_object, keys, data_args) = init_class_kwargs(self, value, USER_FIELDS, UserFieldProps, USERS_COLLECTION, ['id','uid'], **kwargs)

        super().__init__(**data_args);
        for key in keys:
            if key in cls_object:
                if key == "phoneInfo":
                    if bool(cls_object[key]) is False:
                        setattr(self, key, None)
                    else:
                        # print('is phone info', cls_object[key])
                        setattr(self, key, PhoneInfo.from_dict(cls_object[key]))
                else:
                    # print('set attr for', key)
                    setattr(self, key, cls_object[key]);
            else:
                setattr(self, key, None)
                
    


    @staticmethod
    def generate_model(_key_="default_value"):
        user = {};
        props = UserFieldProps
        for k in DictUtils.get_keys(props):
            user[k] = props[k][_key_];
        return user;

    @staticmethod
    def from_dict(user: Any=None, db=None, collection_name=USERS_COLLECTION) -> 'User':
        cls_object, keys = get_object_model_class(user, User, UserFieldProps);
        # print('from dict ===>', cls_object)
        if "firstName" in user and ('first_name' not in cls_object or bool(cls_object['first_name']) is False):
            cls_object["first_name"] = user["firstName"]
        if "lastName" in user and ('last_name' not in cls_object or bool(cls_object['last_name']) is False):
            cls_object["last_name"] = user["lastName"]
            
        _user = User(cls_object, db=db, collection_name=collection_name)
        return _user
    
    def to_dict(self, fields=None):
        # base = super().to_dict(fields)
        # print('base', base['ages']);
        if(getattr(self, 'telephone', None) is None):
            return {}
        telephone = getattr(self, 'telephone', None);
        postal = getattr(self, 'postalCode', None);
        provider = getattr(self, 'provider', None);
        phoneInfo = None;
        firstName = getattr(self, 'first_name', None);
        lastName = getattr(self, 'last_name', None);
        if getattr(self, 'phoneInfo', None) is not None:
            phoneInfo = self.phoneInfo.to_dict();
        if bool(provider) is False:
            provider = "https://adafri.com"
        if phoneInfo is not None and (bool(telephone) is False or telephone!=phoneInfo['number'].strip().replace(' ', '')):
            # print('f self', self)
            # print('f', str(self.phoneInfo))
            telephone = str(phoneInfo['number']).strip().replace(' ', '');
        if bool(postal) is False and bool(self.postal) is True:
            postal = self.postal
        
        # if getattr(self, 'firstName', None) is not None or bool(firstName) is False:
        #     if bool(self.first_name) is True:
        #         firstName = self.first_name
        # if getattr(self, 'lastName', None) is not None or bool(lastName) is False:
        #     if bool(self.last_name) is True:
        #         lastName = self.last_name
        values = {
            "uid": self.uid,
            "email": self.email,
            "password": self.password,
            "displayName": self.displayName,
            "first_name": firstName,
            "last_name": lastName,
            "provider": provider,
            "entrepriseName": self.entrepriseName,
            "entrepriseUrl": self.entrepriseUrl,
            "isConnectWithMailAndPassword": self.isConnectWithMailAndPassword,
            "addresse": self.addresse,
            "photoURL": self.photoURL,
            "profileCompleted": self.profileCompleted,
            "telephone": telephone,
            "postalCode": postal,
            "user_type": self.user_type,
            "authorizedPush": self.authorizedPush,
            "hasApprouvedPolicy": self.hasApprouvedPolicy,
            "emailValidationSendDate": self._emailValidationSendDate,
            "status": self.status,
            "businessType": self.businessType,
            "phoneVerified": self.phoneVerified,
            "emailVerified": self.emailVerified,
            "showPushToken": self.showPushToken,
            "status": self.status,
            "phoneInfo": phoneInfo
        };
        return DictUtils.remove_none_values(values)
    def to_json(self, _fields=None):
        return self.to_dict(_fields)
    
    def get(self):
        doc = self.document_reference().get();
        if doc.exists is False:
            print('user not fon=und', self.uid)
            return None;
        fields = USER_FIELDS
        fields.append('firstName')
        fields.append('lastName')
        d = pydash.pick(doc.to_dict(), fields)
        new_user = User.from_dict(user=d, db=self.db, collection_name=self.collection_name)
        return new_user;

    def query(self, query_params: list, first=False, limit=None):
        result = [];
        query_result = self.custom_query(query_params, first=first, limit=limit, collection_name=self.collection_name)
        # print('query result', query_result)
        if bool(query_result):
            if first:
                return User.from_dict(user=query_result, db=self.db, collection_name=self.collection_name)
            else:
                for doc in query_result:
                    result.append(User.from_dict(user=doc, db=self.db, collection_name=self.collection_name))
                return result
        if first:
                return None
        return [];
    def get_user_id(self):
        return self.uid
    def create(self, **kwargs):
        data_dict = DictUtils.pick_fields(kwargs, UserFields(fields_props=UserFieldProps).filtered_keys('mutable', True));
        user_model = User().from_dict(DictUtils.merge_dict(data_dict, User.generate_model()));
        if bool(user_model.to_json()) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Empty request","INVALID_REQUEST", 1))
        
        email = user_model.email
        password = user_model.password
        
        if bool(email) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Email required","INVALID_REQUEST", 1));
    
        if bool(password) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Password required","INVALID_REQUEST", 1));
        docs = User().collection().where('email', '==', email).get();
        found_users = [];
        if len(docs) > 0:
            for doc in docs:
                found_users.append(User.from_dict(doc.to_dict()).to_json());
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"User wiith email {user_model.email} already exist","INVALID_REQUEST", 1));
    
    
        try:
            user_model.isConnectWithMailAndPassword = bool(user_model.uid) is False
            if bool(user_model.uid) is False:
                user_record = create_firebase_user(email, password);
                user_model.uid = user_record.uid;
                hashed_password = Crypto().encrypt(password);
                user_model.password = hashed_password;
            keys = DictUtils.get_keys(data_dict);
            keys.append('uid');
            User().collection().document(user_model.uid).set({**user_model.to_json(), "createdAt": getTimestamp()}, merge=True)
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, DictUtils.filter_by_fields(user_model.to_json(), keys), None);
        except Exception as e:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(str(e),"INVALID_REQUEST", 1));
    
    def update(self, data):
        try:
            last_value = self.to_dict();
            filtered_value = pydash.pick(data, UserFields(fields_props=UserFieldProps).filtered_keys('editable', True));
            new_value = DictUtils.merge_dict(filtered_value, self.to_dict());
            changed_fields = DictUtils.get_changed_field(last_value, new_value);
            # print('last', last_value)
            # print('new', new_value)
            # print('changed', changed_fields)
            # print('data', data)
            data_update = DictUtils.dict_from_keys(filtered_value, changed_fields);
            if bool(data_update) is False:
                return None;
            self.document_reference().set(data_update, merge=True)
            return DictUtils.dict_from_keys(self.get().to_dict(), changed_fields);
        except Exception as e:
            print(e)
            return None;

    def remove(self, only_mark_as_removed=True):
        try:
            if self.id is None:
                return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Cannot identify User with id {self.id}","INVALID_REQUEST", 1));
            if only_mark_as_removed:
                self.document_reference().set({"is_removed": True})
            else:
                self.document_reference().delete();
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, {"message": f"User {self.id} deleted"}, None);
        except:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"An error occurated while removing authorization code with id {self.id}","INVALID_REQUEST", 1));

    
    def get_firebase_user(self, key: str, value: str) -> 'admin_auth.UserRecord':
        try:
            if key == 'email':
                return admin_auth.get_user_by_email(value)
            elif key == 'uid':
                return admin_auth.get_user(value)
            elif key == 'phone':
                return admin_auth.get_user_by_phone_number(value)
        except ValueError as e:
            raise e
        except FirebaseError as e:
            raise e;    
        except admin_auth.UserNotFoundError as e:
            raise e;  





from ....base.firebase_collection import FirebaseCollectionBase
from .....utils import (DictUtils, init_class_kwargs, get_request_fields)
from .base_adgroup_fields import BaseAdGroupFieldsProps, STANDARD_FIELDS, BASE_ADGROUP_PICKABLE_FIELDS
from typing import Any
from dataclasses import dataclass

from ..ages import Ages
from ..genders import Genders
from ..devices import Devices


def filter_request_fields(fields, default_fields = BASE_ADGROUP_PICKABLE_FIELDS):
    request_fields = get_request_fields(fields, default_fields, [default_fields[0]])
    if request_fields is None or bool(request_fields) is False:
        request_fields = [default_fields[0]]
    return request_fields



@dataclass(init=False)
class BaseAdGroup(FirebaseCollectionBase):
    id: str
    campaign_id: int
    name: str
    status: str
    budget: int
    budgetId: int
    bid: float
    ages: list[Ages]
    genders: list[Genders]
    devicesTargeted: list[Devices]
    devicesExcluded: list[Devices]
    is_removed: bool
    isEnded: bool
    ad_group_id: int
    owner: str
    provider: str
    clientCustomerId: str
    __baseFields: BaseAdGroupFieldsProps

    def __init__(self, adgroup=None, collection_name=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, adgroup, STANDARD_FIELDS, BaseAdGroupFieldsProps, collection_name, ['id'], **kwargs)
        super().__init__(**data_args);
        for key in keys:
            if key == "ages":
                setattr(self, key, Ages().fromListDict(cls_object[key], Ages))
            elif key == "genders":
                setattr(self, key, Genders().fromListDict(cls_object[key], Genders))
            elif key == "devicesTargeted":
                setattr(self, key, Devices().fromListDict(cls_object[key], Devices))
            elif key == "devicesExcluded":
                setattr(self, key, Devices().fromListDict(cls_object[key], Devices))
            else:
                setattr(self, key, cls_object[key]) 


    @staticmethod
    def generate_model(_key_="default_value"):
        adgroup = {};
        props = BaseAdGroupFieldsProps
        for k in DictUtils.get_keys(props):
            adgroup[k] = props[k][_key_];
        return adgroup;

    
    def to_dict(self, fields=None):
        return {
            "id": self.id,
            "campaign_id": self.campaign_id,
            "name": self.name,
            "status": self.status,
            "budget": self.budget,
            "budgetId": self.budgetId,
            "bid": self.bid,
            "ages": Ages().toListDict(self.ages, None),
            "genders": Genders().toListDict(self.genders, None),
            "devicesTargeted": Devices().toListDict(self.devicesTargeted, None),
            "devicesExcluded": Devices().toListDict(self.devicesExcluded, None),
            "is_removed": self.is_removed,
            "ad_group_id": self.ad_group_id,
            "owner": self.owner,
            "provider": self.provider,
            "clientCustomerId": self.clientCustomerId
        }
        


from .adgroup import (AdGroupDisplay, filter_adgroup_display_request_fields)
from .adgroup_fields import (AdGroupDisplayFields, AdGroupDisplayFieldsProps, ADGROUP_DISPLAY_PICKABLE_FIELDS, DISPLAY_ADGROUPS_COLLECTION)



from dataclasses import dataclass
from typing import Any
from  ....utils.utils import DictUtils, BaseFieldsClass,init_class_kwargs
import os

@dataclass
class BaseAdgroupFields(BaseFieldsClass):
    id = "id"
    name = "name"
    campaign_id = "campaign_id"
    ad_group_id = "ad_group_id"
    status = "status"
    ages = "ages"
    genders = "genders"
    clientCustomerId = "clientCustomerId"
    servingStatus = "servingStatus"
    owner = "owner"
    accountId = "accountId"
    createdAt = "createdAt"
    createdBy = "createdBy"
    devicesTargeted = "devicesTargeted"
    devicesExcluded = "devicesExcluded"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



BaseAdgroupFieldsProps = {
    BaseAdgroupFields.id: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    BaseAdgroupFields.name: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    BaseAdgroupFields.status: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": 0
    },
    BaseAdgroupFields.owner: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    BaseAdgroupFields.accountId: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    BaseAdgroupFields.campaign_id: {
        "type": int,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": 0
    },
    BaseAdgroupFields.ad_group_id: {
        "type": int,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": 0
    },
    BaseAdgroupFields.ages: {
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    BaseAdgroupFields.genders: {
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    BaseAdgroupFields.clientCustomerId: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    BaseAdgroupFields.createdBy: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    BaseAdgroupFields.createdAt: {
        "type": (int or str),
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    }
}


STANDARD_FIELDS = BaseAdgroupFields(fields_props=BaseAdgroupFieldsProps).filtered_keys('pickable', True)
BASE_ADGROUP_PICKABLE_FIELDS = BaseAdgroupFields(fields_props=BaseAdgroupFieldsProps).filtered_keys('pickable', True)



from dataclasses import dataclass
from typing import Any
from  ....utils.utils import DictUtils, BaseFieldsClass, BaseClass,init_class_kwargs, get_object_model_class
from .text import AssetText
from .videos import VideoAsset
from .images import AssetNative
@dataclass
class ResponsiveDisplayAdsToPublishFields(BaseFieldsClass):
    id = "id"
    titles = "titles"
    descriptions = "descriptions"
    longHeadline = "longHeadline"
    brand = "brand"
    videosAssets = "videosAssets"
    marketingImages = "marketingImages"
    squareMarketingImages = "squareMarketingImages"
    logoImages = "logoImages"
    landscapeLogoImages = "landscapeLogoImages"
    is_removed = "is_removed"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


@dataclass(init=False)
class ResponsiveDisplayAdsToPublish(BaseClass):
    id: str
    titles: list[AssetText]
    descriptions: list[AssetText]
    longHeadline: str
    brand: AssetText
    videosAssets: list[VideoAsset]
    marketingImages: list[AssetNative]
    squareMarketingImages: list[AssetNative]
    logoImages: list[AssetNative]
    landscapeLogoImages: list[AssetNative]
    is_removed: bool

    def __init__(self, asset=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, asset, None, None, None, [], **kwargs)
        for key in keys:
            if key=='titles':
                setattr(self, key, AssetText().fromListDict(cls_object[key]))
            elif key=='descriptions':
                setattr(self, key, AssetText().fromListDict(cls_object[key]))
            elif key=='brand':
                setattr(self, key, AssetText().from_dict(cls_object[key]))
            elif key=='videosAssets':
                setattr(self, key, VideoAsset().fromListDict(cls_object[key]))
            elif key=='marketingImages':
                setattr(self, key, AssetNative().fromListDict(cls_object[key]))
            elif key=='squareMarketingImages':
                setattr(self, key, AssetNative().fromListDict(cls_object[key]))
            elif key=='logoImages':
                setattr(self, key, AssetNative().fromListDict(cls_object[key]))
            elif key=='landscapeLogoImages':
                setattr(self, key, AssetNative().fromListDict(cls_object[key]))
            else:
                setattr(self, key, cls_object[key]) 
    def to_dict(self, field=None):
        return DictUtils.remove_none_values({
            "id": self.id,
            "titles": AssetText().toListDict(self.titles, None),
            "descriptions": AssetText().toListDict(self.descriptions, None),
            "longHeadline": self.longHeadline,
            "brand": self.brand.to_dict(None),
            "videosAssets": VideoAsset().toListDict(self.videosAssets, None),
            "marketingImages": AssetNative().toListDict(self.marketingImages, None),
            "squareMarketingImages": AssetNative().toListDict(self.squareMarketingImages, None),
            "logoImages": AssetNative().toListDict(self.logoImages, None),
            "landscapeLogoImages": AssetNative().toListDict(self.landscapeLogoImages, None),
            "is_removed": self.is_removed
        })
    @staticmethod
    def from_dict(obj: Any) -> 'ResponsiveDisplayAdsToPublish':
        cls_object, keys = get_object_model_class(obj, ResponsiveDisplayAdsToPublish, None);
        return ResponsiveDisplayAdsToPublish(cls_object)


@dataclass(init=False)
class DisplayAdsToPublishFields(BaseFieldsClass):
    id = "id"
    name = "name"
    mediaId = "mediaId"
    type = "type"
    width = "width"
    height = "height"
    urls = "urls"
    previewUrl = "previewUrl"
    resourceName = "resourceName"
    sourceUrl = "sourceUrl"
    type = "type"
    mimeType = "mimeType"
    owner = "owner"
    fileSize = "fileSize"
    clientCustomerId = "clientCustomerId"

@dataclass(init=False)
class DisplayAdsToPublish(BaseClass):
    id: str
    name: str
    mediaId: int
    type: str
    width: str
    height: str
    urls: str
    previewUrl: str
    resourceName: str
    sourceUrl: str
    type: str
    mimeType: str
    owner: str
    fileSize: str
    clientCustomerId: str
    def __init__(self, ad=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, ad, None, None, None, [], **kwargs)
        for key in keys:
            setattr(self, key, cls_object[key]) 
    def to_dict(self, field=None):
        return DictUtils.remove_none_values({
            "id": self.id,
            "name": self.name,
            "mediaId": self.mediaId,
            "type": self.type,
            "width": self.width,
            "height": self.height,
            "urls": self.urls,
            "previewUrl": self.previewUrl,
            "resourceName": self.resourceName,
            "sourceUrl": self.sourceUrl,
            "type": self.type,
            "mimeType": self.mimeType,
            "owner": self.owner,
            "fileSize": self.fileSize,
            "clientCustomerId": self.clientCustomerId
        })
    @staticmethod
    def from_dict(obj: Any) -> 'DisplayAdsToPublish':
        cls_object, keys = get_object_model_class(obj, DisplayAdsToPublish, None);
        return DisplayAdsToPublish(cls_object)





from dataclasses import dataclass
from  .....utils.utils import BaseFieldsClass


@dataclass
class BaseAdGroupFields(BaseFieldsClass):
    id = "id"
    campaign_id = "campaign_id"
    ad_group_id = "ad_group_id"
    name = "name"
    status = "status"
    budget = "budget"
    owner = "owner"
    createdAt = "createdAt"
    createdBy = "createdBy"
    accountId = "accountId"
    is_removed = "is_removed"
    ages = "ages"
    genders = "genders"
    devicesTargeted = "devicesTargeted"
    devicesExcluded = "devicesExcluded"
    clientCustomerId = "clientCustomerId"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)



BaseAdGroupFieldsProps = {
    BaseAdGroupFields.id: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    BaseAdGroupFields.name: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    BaseAdGroupFields.status: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": ''
    },
    BaseAdGroupFields.budget: {
        "type": (int or float),
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": 0
    },
    BaseAdGroupFields.owner: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    BaseAdGroupFields.accountId: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    BaseAdGroupFields.campaign_id: {
        "type": int,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": 0
    },
    BaseAdGroupFields.ad_group_id: {
        "type": int,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": 0
    },
    BaseAdGroupFields.is_removed: {
        "type": bool,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": False
    },
    BaseAdGroupFields.ages: {
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    BaseAdGroupFields.genders: {
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    BaseAdGroupFields.clientCustomerId: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    BaseAdGroupFields.createdBy: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    BaseAdGroupFields.createdAt: {
        "type": (int or str),
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    BaseAdGroupFields.devicesTargeted: {
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    BaseAdGroupFields.devicesExcluded: {
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    }
}


STANDARD_FIELDS = BaseAdGroupFields(fields_props=BaseAdGroupFieldsProps).filtered_keys('pickable', True)
BASE_ADGROUP_PICKABLE_FIELDS = BaseAdGroupFields(fields_props=BaseAdGroupFieldsProps).filtered_keys('pickable', True)




from .adgroups import BaseAdGroup, BaseAdGroupFields, BaseAdGroupFieldsProps, BASE_ADGROUP_PICKABLE_FIELDS
from .client_customer_id import ClientCustomerId
from .client_customer_id_fields import CLIENT_CUSTOMER_ID_PICKABLE_FIELDS, ClientCustomerIdFields, ClientCustomerIdFieldsProps


from .oauth import (client, client_fields)
from .firebase_auth import (firebase_auth, FirestoreApp)
def handleauth():
    return "handled"
# from .oauth.models.client_fields import (ClientFieldProps, ClientFields)
# from .models.client import (Client)
# from .firebase import FirestoreApp


genders_data = [
    { "criterionId": 10, "name": 'Homme', "text": 'Male' },
    { "criterionId": 11, "name": 'Femme', "text": 'Female' },
    {
        "criterionId": 20,
        "name": 'Non renseignés à Google',
        "text": 'Undetermined',
        "info": '(il est recommandé de laisser cette option cochée)'
    }
]

ages_data = [
    { "criterionId": 503001, "name": '18-24 ans', "text": '18-24 yo' },
    { "criterionId": 503002, "name": '25-34 ans', "text": '25-34 yo' },
    { "criterionId": 503003, "name": '35-44 ans', "text": '34-44 yo' },
    { "criterionId": 503004, "name": '45-54 ans', "text": '45-54 yo' },
    { "criterionId": 503005, "name": '55-64 ans', "text": '55-64 yo' },
    { "criterionId": 503006, "name": '+64 ans', "text": '64up' },
    {
        "criterionId": 503999,
        "name": 'Non renseignés à Google',
        "text": 'Undetermined',
        "info": '(il est recommandé de laisser cette option cochée)'
    }
]


from .firebase_auth import (FirestoreApp, create_firebase_user, create_firebase_custom_token, set_custom_claims,
                            generate_email_verification_link, generate_password_reset_link, update_firebase_user,
                            get_user_by_uid, get_user_by_email, get_user_by_phone_number)


from .adgroups import (AdGroupDisplay, AdGroupDisplayFields, AdGroupDisplayFieldsProps, filter_adgroup_display_request_fields, ADGROUP_DISPLAY_PICKABLE_FIELDS)
from .models import (
    CampaignDisplay, CampaignDisplayFields, 
    CampaignDisplayFieldsProps, filter_campaign_display_request_fields,
    CAMPAIGN_DISPLAY_PICKABLE_FIELDS, CAMPAIGN_DISPLAY_NOT_PICKABLE_FIELDS
    )



from dataclasses import dataclass
from typing import Any
from  ....utils.utils import DictUtils, BaseFieldsClass, BaseClass,init_class_kwargs,get_object_model_class
from ...base.firebase_collection import FirebaseCollectionBase

@dataclass
class UserInterestFields(BaseFieldsClass):
    id = "id"
    name = "name"
    type = "type"
    userInterestParentId = "userInterestParentId"
    criterionId = "criterionId"
    criterionType = "criterionType"
    isTargeted = "isTargeted"
    isExcluded = "isExcluded"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


UserInterestFieldsProps = {
    UserInterestFields.id: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    UserInterestFields.name: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    UserInterestFields.type: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    UserInterestFields.criterionId: {
        "type": str or int or float,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    UserInterestFields.criterionType: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    UserInterestFields.userInterestParentId: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    UserInterestFields.isTargeted: {
        "type": bool,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    UserInterestFields.isExcluded: {
        "type": bool,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    }
}

@dataclass(init=False)
class UserInterest(FirebaseCollectionBase):
    id: str
    name: str
    type: str
    criterionType: str
    userInterestParentId: str
    criterionId: int
    isTrageted: bool
    isExcluded: bool
    __baseFields: UserInterestFieldsProps
    def __init__(self, user_interest=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, user_interest, None, None, None, [], **kwargs)
        super().__init__(**data_args);
        for key in keys:
            setattr(self, key, cls_object[key]) 
    def to_dict(self, field=None):
        return DictUtils.remove_none_values({
            "id": self.id,
            "name": self.name,
            "type": self.type,
            "criterionType": self.criterionType,
            "userInterestParentId": self.userInterestParentId,
            "criterionId": self.criterionId,
            "isTrageted": self.isTrageted,
            "isExcluded": self.isExcluded
        })
    @staticmethod
    def from_dict(user_interest = None, db=None, collection_name=None) -> 'UserInterest':
        cls_object, keys = get_object_model_class(user_interest, UserInterest, None);
        value  = UserInterest(cls_object);
        if db is not None:
            value.db = db
        if collection_name is not None:
            value.collection_name = collection_name
        return value
    
    def loadBrand(self):
        brands = self.query(UserInterest, "user_interest", [], False, None, "BRAND-user-interest-fr")
        if bool(brands) is True:
            return self.toListDict(brands, None);
        return []
    def loadInMarket(self):
        in_market = self.query(UserInterest, "user_interest", [], False, None, "IN_MARKET-user-interest-fr")
        if bool(in_market) is True:
            return self.toListDict(in_market, None);
        return []
    def loadMobileAppInstaller(self):
        mobile_app = self.query(UserInterest, "user_interest", [], False, None, "MOBILE_APP_INSTALL_USER-user-interest-fr")
        if bool(mobile_app) is True:
            return self.toListDict(mobile_app, None);
        return []
    def load(self):
        try:
            result = {};
            brands = self.loadBrand();
            if bool(brands) is True:
                result['brand'] = brands
            in_market = self.loadInMarket();
            if bool(in_market) is True:
                result['in_market'] = in_market
            mobile_app = self.loadMobileAppInstaller();
            if bool(mobile_app) is True:
                result['mobile_app_installer'] = mobile_app
            return result
        except Exception as e:
            print(str(e))
            return None;





from .models import client
from .models import client_fields
from .models.client_fields import (ClientFieldProps, ClientFields)
from .models.client import (OAuthClient)
from .models import grant
from .models import grant_fields
from .models.grant_fields import (GrantFieldsProps, GrantFields)
from .models.grant import (OAuthGrant, AuthorizationCodeGrant, OpenIDCode, OpenIDImplicitGrant, OpenIDHybridGrant, OpenIDAuthorizationCodeGrant)
from .models import token
from .models import token_fields
from .models.token_fields import (TokenFieldsProps, TokenFields)
from .models.token import (OAuthToken, TokenGenerator, TokenValidator, RefreshTokenGrant, TokenRevocationEndpoint, JwtTokenGenerator, JwtTokenValidator, RefreshTokenGrant)
from .models import code
from .models import code_fields
from .models.code import (Code)
from .models.code_fields import (CodeFieldsProps, CodeFields)
from .token_debug import check_token, is_existing_token
# from .key_handler import public_jwk
from .server import (oidc_authorization_server, authorization_server, require_oauth, require_oidc_oauth, config_oauth, config_oidc_oauth, guards_oidc)


from dataclasses import dataclass
from typing import Any
from  ....utils.utils import DictUtils, BaseFieldsClass, BaseClass,init_class_kwargs,get_object_model_class,get_request_fields


@dataclass
class LocationFields(BaseFieldsClass):
    id = "id"
    locationName = "locationName"
    displayType = "displayType"
    targetingStatus = "targetingStatus"
    reach = "reach"
    canonicalName = "canonicalName"
    criterionId = "criterionId"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.class_fields_props = LocationFieldsProps


LocationFieldsProps = {
    LocationFields.id: {
        "type": int,
        "required": True,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    LocationFields.locationName: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    LocationFields.canonicalName: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    LocationFields.displayType: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    LocationFields.displayType: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    LocationFields.targetingStatus: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    LocationFields.reach: {
        "type": int,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    }
}

@dataclass(init=False)
class Location(BaseClass):
    id: str
    locationName: str
    displayType: str
    targetingStatus: str
    reach: str
    canonicalName: str
    criterionId: int
    def __init__(self, location=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, location, None, None, None, [], **kwargs)
        for key in keys:
            setattr(self, key, cls_object[key]) 
    def to_dict(self, field=None):
        return DictUtils.remove_none_values({
            "id": self.id,
            "locationName": self.locationName,
            "targetingStatus": self.targetingStatus,
            "reach": self.reach,
            "canonicalName": self.canonicalName,
            "criterionId": self.criterionId
        })
    @staticmethod
    def from_dict(obj: Any) -> 'Location':
        cls_object, keys = get_object_model_class(obj, Location, None);
        return Location(cls_object)
    
def filter_locations_request_fields(fields, default_fields = LocationFields().filtered_keys('pickable')):
    request_fields = get_request_fields(fields, default_fields, None)
    if request_fields is None or bool(request_fields) is False:
        request_fields = [default_fields[0]]
    return request_fields


import json
import pydash
from typing import List, Iterable, Any, Union
import hashlib
from cryptography.fernet import Fernet
import os
import base64
from datetime import (date, datetime, time)
from urllib.parse import urlparse, parse_qs
import re
import random
from google.cloud.firestore_v1.document import Timestamp
from google.api_core.datetime_helpers import DatetimeWithNanoseconds
from datetime import timedelta

hash = hashlib.sha1(str(os.getenv('CRYPTO_KEY')).encode())
ENCRYPTION_KEY = base64.b64encode(hash.hexdigest()[:32].encode()).decode();
from urllib.parse import unquote

camel_pat = re.compile(r'([A-Z])')
under_pat = re.compile(r'_([a-z])')


def interger(data):
    if type(data) is int:
        return data;
    if type(data) is str:
        try:
            return int(data);
        except Exception:
            return None;
    return None;
def boolean(data):
    if type(data) is bool:
        return data;
    if type(data) is str:
        if data.lower() == 'true':
            return True;
        if data.lower() == 'false':
            return False;
    return None;

class Object:
    def __init__(self, **kwargs):
        if kwargs is not None:
            for k, v in kwargs.items():
                setattr(self, k, v);

def split_by_crlf(s):
    return [v for v in s.splitlines() if v]

def split_by_comma(s):
    sp = s.split(',')
    if len(sp) > 0:
        splitted = [];
        for split in sp:
            splitted.append(str(split).lstrip())
        return splitted;
    return []


def read_file(_filename=None, isJson=False):
    filename = _filename
    fp = filename
    if filename is None:
        return None
    if os.path.isabs(filename) is False:
        fp = os.getcwd() + "/" + filename
    content = open(fp, 'r').read()
    try:
        if isJson is True:
            return json.loads(content)
        else:
            return content
    except Exception as e:
        return None

def isBase64(data):
    try:
        decoded = decode_base64(data)
        return decoded != data
    except Exception:
            return False

def encode_if_not_base64(data) -> 'str | None':
    try:
        if isBase64(data):
            return data
        return encode_base64(data)
    except Exception as e:
        return None;
 
def decode_if_base64(data) -> 'str | None':
    try:
        if isBase64(data):
            return decode_base64(data)
        return data;
    except Exception as e:
        print('base 64 decode error', e)
        return None

def generate_random_code():
    start = random.randint(2000, 5000)
    start1 = random.randint(6000, 9000)
    end = random.randint(start, random.randint(start, start1))
    return random.randint(start, end)

def format_query_filter(key: str, value: any, comparator: str):
    """
        key A key existing in document data
        @{value} The value used to compare
        comparator The method used to compare (==, in...)
    """
    query = {};
    query["key"] = key
    query["value"] = value
    query['comp'] = comparator
    return query

def camel_to_underscore(name):
    return camel_pat.sub(lambda x: '_' + x.group(1).lower(), name)

def underscore_to_camel(name):
    return under_pat.sub(lambda x: x.group(1).upper(), name)

def convert_to_camelcase(obj):
    data = {}
    for key in pydash.keys(obj):
        data[underscore_to_camel(key)] = obj[key];
    return data;
def convert_to_underscore(obj):
    data = {}
    for key in pydash.keys(obj):
        data[camel_to_underscore(key)] = obj[key];
    return data;

def add_param_to_url(input_url,  params):
    import urllib.parse as urlparse
    from urllib.parse import urlencode
    url_parts = list(urlparse.urlparse(input_url))
    query = dict(urlparse.parse_qsl(url_parts[4]))
    query.update(params)
    url_parts[4] = urlencode(query)
    return urlparse.urlunparse(url_parts)

def get_url_params(url):
    try:
        parse_result = urlparse(url)
        query_params_dict = parse_qs(parse_result.query);
        query_params = {};
        for key in pydash.keys(query_params_dict):
            query_params[key] = ' '.join(map(str, query_params_dict[key]));
        return query_params;
    except:
        return None;


def encode_base64(data: str):
    string_bytes = data.encode("ascii")
    base64_bytes = base64.b64encode(string_bytes)
    base64_string = base64_bytes.decode("ascii")
    return base64_string;

def decode_base64(data_base64: str):
    base64_bytes = data_base64.encode("ascii")
    string_bytes = base64.b64decode(base64_bytes)
    decoded_string = string_bytes.decode("ascii")
    return unquote(decoded_string)

class Crypto:

    
    # print(hash.hexdigest())
    fernet = Fernet(ENCRYPTION_KEY);
    def encrypt(self, message:str):
        return self.fernet.encrypt(message.encode()).decode();

    def decrypt(self, encrypted:str):
        return self.fernet.decrypt(encrypted).decode();

    def hash(self, message:str):
        return hashlib.sha256(str(message).encode()).hexdigest();
    
    def generate_id(self, message:str):
        return hashlib.md5(str(message).encode()).hexdigest();

    def generate_token(self, message:str):
        encrypted_token = self.encrypt("token~"+message);
        hash_token = hashlib.md5(str(encrypted_token).encode()).hexdigest();
        return encode_base64(hash_token)
    
    def decrypt_token(self, message:str):
        try:
            d = self.decrypt(decode_base64(message));
            token_split = d.split('~');
            token = json.loads(token_split[1])
            return token
        except:
            return None;


class RequestFields:
    data = 'data';
    fields = 'fields';

def get_request_fields(field_string, fields, default_fields):
    if type(field_string) is list:
        return field_string
    
    request_fields =[];
    load_fields = [];
    if type(field_string) is list:
        load_fields = field_string
    else:
        load_fields = ArrayUtils.join_string_to_array(field_string,',');
    # print('load_fields', load_fields)
    # print('load_fields', fields)
    hasObjective = False
    if bool(load_fields):
        for f in load_fields:
            parts = parts = f.split('.')
            attr = parts[0]
            if attr == 'objective':
                # print('attr', attr)
                # print('parts', parts)
                # print('fields', fields)
                hasObjective = True
            if attr in fields:
                request_fields.append(f)
            # if len(parts) > 1:
            #     if attr in fields:
            # else:
            #     if attr in:
            #         request_fields.append(f)

        # request_fields = ArrayUtils.pick(fields, load_fields);
    
    if bool(request_fields) is False:
        request_fields = [fields[0]]
    if default_fields is not None:
        for f in default_fields:
            if f not in request_fields:
                request_fields.append(f);
    
    response = [];
    for f in request_fields:
        response.append(str(f).strip())
    
    return response;

import inspect
def get_class_properties(cls, property='__match_args__'):
        inspected = inspect.getmembers(cls, lambda a:not(inspect.isroutine(a)))
        # attr = [a for a in inspected if not(a[0].startswith('__') and a[0].endswith('__'))]
        attr = [a[1] for a in inspected if a[0]=='__match_args__']
        if(len(attr)>0):
            if type(attr[0]) is tuple:
                return list(attr[0])
        return list(attr)

def get_object_model_class(object_model, cls, fields=None, property='__match_args__'):
        obj = object_model
        if obj is None:
            obj = {};
        attributes = get_class_properties(cls, property);
        for key in attributes:
            if key not in obj:
                try:
                    obj[key] = getattr(cls, key, None)
                except Exception as err:
                    obj[key] = None

        data_object = obj.copy()
        keys = DictUtils.get_keys(data_object);
        if fields is not None:
            for key in keys:
                if key in fields:
                    # print(data_object[key])
                    if key not in data_object or data_object[key] is None or data_object[key] == 'None':
                        if 'default_value' in fields[key]:
                            data_object[key] = fields[key]['default_value']
        return data_object, keys

def init_class_kwargs(cls, obj, class_fields, class_fields_props, class_collection_name, ids_key: list[str], **kwargs):
        cls_object, keys = get_object_model_class(obj, cls, class_fields_props);
        kwargs['fields'] = class_fields
        kwargs['fields_props'] = class_fields_props

        for key in ids_key:
            documentId = None;
            if key in cls_object:
                documentId = cls_object[key]
            if documentId is not None and bool(documentId):
                # print('docId', documentId)
                kwargs['id'] = documentId;
                break;
        
        collection_name = getattr(kwargs, 'collection_name', None)
        if collection_name is None or bool(collection_name) is False:
            if class_collection_name is not None:
                kwargs['collection_name'] = class_collection_name;
        return cls_object, keys, kwargs;

class BaseClass:

    @staticmethod
    def from_dict(element, db=None, collection_name=None) -> 'Any':
        return NotImplementedError('from_json  static method must be implemented')
    def __init__(self, fields=None, fields_props=None):
        self.fields = fields;
        self.fields_props = fields_props;
    def toListDict(self, elements, fields=['id']):
        if not isinstance(elements, list):
            return None;
        result = []
        for element in elements:
            if getattr(self, 'to_dict', None) is not None:
                if fields is not None and type(fields) is list and len(fields)>0:
                    result.append(DictUtils.filter_by_fields(element.to_dict(), fields))
                else:
                    result.append(element.to_dict())
            elif getattr(self, 'to_json', None) is not None:
                result.append(element.to_json(fields))
        return result


    def fromListDict(self, elements, cls=None):
        if elements is None:
            return []
        result = []
        for element in elements:
            if getattr(self, 'from_dict', None) is not None:
                result.append(self.from_dict(element))
        return result
        
    def to_json(self, _fields=None):
        fields = None
        if getattr(self, 'fields', None) is not None:
            fields = self.fields
        else:
            fields = self.__dict__.items()
        if _fields is not None and type(_fields) is list and len(_fields)>0:
            fields = get_request_fields(_fields, fields, [fields[0]]);
            
        json_object, keys = get_object_model_class({}, self, getattr(self, "fields_props", None));
        if fields is None or type(fields) is not list or len(fields)==0:
            return json.loads(json.dumps(json_object, cls=JsonEncoder))
        return json.loads(json.dumps(DictUtils.filter_by_fields(json_object, fields), cls=JsonEncoder, fields=fields))
    
class BaseFieldsClass:
    class_fields_props: dict
    def __init__(self, fields_props=None, **kwargs):
        self.class_fields_props = fields_props
            
    def keys(self):
        return DictUtils.get_keys(self.class_fields_props);

    def filtered_keys(self, field, condition=True):
        mutable = DictUtils.filter(self.class_fields_props, DictUtils.get_keys(self.class_fields_props), field, condition)
        return DictUtils.get_keys(mutable);

    def check_create_fields(self, request_data, fieldsProps = None):
        result = {}
        fields = fieldsProps
        if fieldsProps is None:
            fields = self.class_fields_props
        if not request_data or not isinstance(request_data, dict):
            return {"error": "Invalid request data"}
        
        for field in fields:
            if field in request_data:
                # if isinstance(request_data[field], fields[field]['type']):
                types = fields[field]['type']
                if not isinstance(types,list):
                    types = [types]
                if type(request_data[field]) in types:
                    if fields[field]['mutable']:
                        result[field] = request_data[field]
                else:
                    if fields[field]['required']:
                        return {"error": f"Invalid type for field: {field} {type(request_data[field])}"}
            else:
                if fields[field]['required']:
                    return {"error": f"Missing required field: {field}"}
        return result

    @staticmethod
    def merge_model_data(model, data):
        if not isinstance(data, dict) or not isinstance(model, dict):
            return None
        result = data
        for key in model:
            if key not in data:
                result[key] = model[key]
        return result

class JsonEncoder(json.JSONEncoder):
    def __init__(self, *args, fields=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields = fields
    def default(self, obj):
        # try:
        #     iterable = iter(obj)
        # except TypeError:
        #     pass
        # else:
        #     return list(iterable)
        # Let the base class default method raise the TypeError
        #return json.JSONEncoder.default(self, obj)
 
        attr = getattr(obj, '__dict__', None)
        to_json = getattr(obj, 'to_json', None)
        if isinstance(obj, datetime):
            return obj.isoformat()
        elif hasattr(obj, "to_dict"):
            # return {k: v for k, v in obj.__dict__.items()}
            return obj.to_dict(self.fields)
        elif hasattr(obj, "__dict__"):
            # return {k: v for k, v in obj.__dict__.items()}
            return obj.__dict__
        else:
            return None

        # if to_json is not None:
        #     # return obj.__dict__
        #     print('to json',obj.) 
        #     return obj
        # return None

class DateUtils:

    @staticmethod
    def get_current_date_time():
        return datetime.now();

    @staticmethod
    def reset_date_time(_date: datetime):
        # new_date = _date.replace(hour=0, minute=0, second=0, microsecond=0)
        return datetime.combine(_date.date(), time(0, 0, 0, 0));
    @staticmethod
    def is_past_date(date):
        try:
            now = DateUtils.reset_date_time(datetime.now())
            return now > DateUtils.reset_date_time(date);
        except Exception as e:
            print('e', e)
            return None;
    @staticmethod
    def is_superior(date: datetime, date2: datetime):
        try:
            return DateUtils.reset_date_time(date) > DateUtils.reset_date_time(date2);
        except Exception as e:
            return None;
    @staticmethod
    def is_equal(date: datetime, date2: datetime):
        try:
            return DateUtils.reset_date_time(date) == DateUtils.reset_date_time(date2);
        except Exception as e:
            return None;
    @staticmethod
    def from_iso(date_str):
        try:
            return date.fromisoformat(date_str);
        except Exception as e:
            return None;

    @staticmethod
    def from_timestamp(date_str):
        try:
            return datetime.fromtimestamp(float(date_str));
        except Exception as e:
            return None;
    @staticmethod
    def isExpired(date_str):
        try:
            return datetime.fromisoformat(date_str) < datetime.now();
        except Exception as e:
            return None;
    @staticmethod
    def convert_firestore_timestamp_to_str(firestore_timestamp: Timestamp, fmt: str = "%Y-%m-%d %H:%M:%S") -> str:
        """
        Convert Firestore Timestamp to a formatted string.

        Args:
            firestore_timestamp (Timestamp): The Firestore Timestamp to convert.
            fmt (str): The format for the output string (default is "%Y-%m-%d %H:%M:%S").

        Returns:
            str: The formatted string representation of the timestamp.
        """

        dt = None
        if isinstance(firestore_timestamp, Timestamp) is True:
            # Convert Firestore Timestamp to a datetime object
            dt = firestore_timestamp.to_datetime()
        elif isinstance(firestore_timestamp, DatetimeWithNanoseconds) is True:
            dt = DateUtils.from_timestamp(firestore_timestamp.timestamp())
        
        if dt is None:
            return firestore_timestamp
        
        # Format the datetime object to a string
        return dt.strftime(fmt)
    @staticmethod
    def getDateFromEn(date_string: str):
        slashes = len(str(date_string).split('/')) == 3
        date_string = str(date_string).replace('/', '-')
        try:
            date_format = "%Y-%m-%d"
            date_obj = datetime.strptime(date_string, date_format)
            return {"status": "ok", "date": date_obj, "format": "en", "isSlash": slashes}
        except ValueError as e:
            try:
                date_format = "%d-%m-%Y"
                date_obj = datetime.strptime(date_string, date_format)
                return {"status": "ok", "date": date_obj, "format": "fr", "isSlash": slashes}
            except:
                return {"status": "error", "date": None, "error": {'message': f"Incorrect data format {date_string}, should be YYYY-MM-DD"}}

    @staticmethod
    def format_campaign_date(data: dict, set_end_date = True):
        if not isinstance(data, dict):
            return {"error": {"message":"Invalid data"}}
        start = None
        end = None
        for key, value in data.items():
            if key =='startDate':
                start = value
            elif key == 'endDate':
                end = value
        try:
            _DATE_FORMAT_GOOGLE = "%Y%m%d"
            _DATE_FORMAT_FR = "%d/%m/%Y"
            _DATE_FORMAT_EN = "%Y-%m-%d"
            campaign = {};
            startDate = None;
            endDate = None
            startDateFormat = None;
            endDateFormat = None;
            slashes = []

            if start is not None:
                date_object = DateUtils.getDateFromEn(start);
                if date_object['status']=='error':
                    return {"status": "error", "campaign": None, "error": date_object['error']};
                
                startDate = date_object['date']
                if DateUtils.is_past_date(startDate) is True:
                    return {"status": "error", "campaign": None, "error": {'message': 'Start date cannot be in the past'}};
                startDateFormat = date_object['format']
                slashes.append(date_object['isSlash'])
                campaign['startDateFormattedGoogle'] = datetime.strftime(startDate, _DATE_FORMAT_GOOGLE);
                campaign['startDateFrench'] = datetime.strftime(startDate, _DATE_FORMAT_FR);
                campaign['startDateEnglish'] = datetime.strftime(startDate, _DATE_FORMAT_EN);

            # Optional: Set the end date.
            if startDate is not None and set_end_date is True:
                endDate = startDate + timedelta(weeks=10)
            # end_time = date + timedelta(weeks=10);
            if end is not None and bool(end) is True:
                date_object = DateUtils.getDateFromEn(end);
                if date_object['status']=='error':
                    return {"status": "error", "campaign": None, "error": date_object['error']};
                endDate = date_object['date'];
            
            if endDate is not None:
                if DateUtils.is_past_date(endDate) is True:
                    return {"status": "error", "campaign": None, "error": {'message': 'End date cannot be in the past'}};
                if DateUtils.is_superior(startDate, endDate) is True:
                    return {"status": "error", "campaign": None, "error": {'message': 'Start date cannot be superior to end date'}};
                if DateUtils.is_equal(startDate, endDate) is True:
                    return {"status": "error", "campaign": None, "error": {'message': 'Start date cannot be equal to end date'}};
                endDateFormat = date_object['format']
                campaign['endDateFormattedGoogle'] = datetime.strftime(endDate, _DATE_FORMAT_GOOGLE);
                campaign['endDateFrench'] = datetime.strftime(endDate, _DATE_FORMAT_FR);
                campaign['endDateEnglish'] = datetime.strftime(endDate, _DATE_FORMAT_EN);
                if 'endDate' not in data:
                    format = None
                    if startDateFormat == 'fr' or endDateFormat == 'fr':
                        format = datetime.strftime(endDate, _DATE_FORMAT_FR);
                    else:
                        format = datetime.strftime(endDate, _DATE_FORMAT_EN);
                    if True in slashes:
                        format = format.replace('-', '/')
                    campaign['endDate'] = format;

            

            return {"status": "ok", "campaign": campaign, "error": None};
        except Exception as e:
            print(e);
            return {"status": "error", "campaign": None, "error": {'message': str(e)}};

class DictUtils:

    @staticmethod
    def pick_object_values(obj, keys: list[str]):
        cls_object = {};
        for attr in keys:
            cls_object[attr] = pydash.get(obj, attr, None)
        return cls_object
    @staticmethod
    def every_match(element, exceptions=[]):
        if element is None:
            return False;
    
        data = pydash.omit(element, exceptions);
        keys = pydash.keys(data);
        if bool(keys) == False:
            return False;
    
        for key in keys:
            if bool(data[key]) is False:
                print("key", key);
                print("data", data[key]);
                return False;
        return True;

    @staticmethod
    def difference(first_dict, second_dict):
        return { k : second_dict[k] for k in set(second_dict) - set(first_dict) }

    @staticmethod
    def compare_nested_objects(obj1: dict, obj2: dict, path: str = '') -> dict:
        """
        Compare two nested dictionaries and return the changed fields, including lists.

        Args:
            obj1 (dict): The first dictionary to compare.
            obj2 (dict): The second dictionary to compare.
            path (str): The current path in the nested structure (used for recursion).

        Returns:
            dict: A dictionary containing the paths of changed fields and their values.
        """
        changes = {}

        # Get all keys from both dictionaries
        all_keys = set(obj1.keys()).union(set(obj2.keys()))

        for key in all_keys:
            # Create the current path
            current_path = f"{path}.{key}" if path else key

            # Check if the key exists in both dictionaries
            if key in obj1 and key in obj2:
                if isinstance(obj1[key], dict) and isinstance(obj2[key], dict):
                    # If both values are dictionaries, recurse
                    nested_changes = DictUtils.compare_nested_objects(obj1[key], obj2[key], current_path)
                    changes.update(nested_changes)
                elif isinstance(obj1[key], list) and isinstance(obj2[key], list):
                    # If both values are lists, compare them
                    list_changes = ArrayUtils.compare_lists(obj1[key], obj2[key], current_path)
                    changes.update(list_changes)
                elif obj1[key] != obj2[key]:
                    # If values are different, record the change
                    changes[current_path] = {'old_value': obj1[key], 'new_value': obj2[key]}
            elif key in obj1:
                # Key exists only in obj1
                changes[current_path] = {'old_value': obj1[key], 'new_value': None}
            else:
                # Key exists only in obj2
                changes[current_path] = {'old_value': None, 'new_value': obj2[key]}

        return changes



    @staticmethod
    def get_changed_field(a, b):

        keys_a = pydash.keys(a);
        keys_b = pydash.keys(b);
        changed_a = [];
        changed_b = [];
        if isinstance(a, list):
            print('a', a)
            print('b', b)
            if isinstance(b, list):
                for i in range(len(a)):
                    if isinstance(a[i], dict):
                        changed = DictUtils.get_changed_field(a[i], b[i]);
                        if len(changed)>0:
                            changed_a.append(a[i]);
                            continue;
                    if a[i]!=b[i]:
                        changed_a.append(a[i]);
            else:
                changed_a.append(a);
            
            for i in range(len(b)):
                if isinstance(b[i], dict):
                    changed = DictUtils.get_changed_field(b[i], a[i]);
                    if len(changed)>0:
                        changed_b.append(i);
                        continue;
                if b[i]!=a[i]:
                    changed_b.append(i);
            changed_fields = pydash.uniq(changed_a + changed_b);
            return changed_fields;

        for k in keys_a:
            if k in b:
                if isinstance(b[k], dict):
                    changed = DictUtils.get_changed_field(b[k], a[k]);
                    if k == "phoneInfo":
                        print('changed e', changed)
                    if len(changed)>0:
                        changed_a.append(k);
                        continue;
                # print('a', a)
                # print('b', b)
                # print('k', k)
                # print('k', {a: a[k], b: b[k]})
                if a[k]!=b[k]:
                    changed_a.append(k);
            else:
                changed_a.append(k);
        
        for k in keys_b:
            if k in a:
                if b[k]!=a[k]:
                    if isinstance(a[k], dict):
                        changed = DictUtils.get_changed_field(a[k], b[k]);
                        if len(changed)>0:
                            changed_b.append(k);
                            continue;
                    changed_b.append(k);
            else:
                changed_b.append(k);
        
        changed_fields = pydash.uniq(changed_a + changed_b);
        return changed_fields;

    @staticmethod
    def dict_from_keys(a, keys):
        return pydash.pick(a, keys);

    @staticmethod
    def get_keys(a):
        return pydash.keys(a);
    
    @staticmethod
    def merge_dict(current_dict, target_dict, force=False):
        final_dict = target_dict.copy()
        # print('target_dict', target_dict)
        for key in pydash.keys(current_dict):
            if force is True or key not in final_dict or final_dict[key]!=current_dict[key]:
                final_dict[key] = current_dict[key]
        return final_dict;

    @staticmethod
    def string_to_json(element):
        try:
            return json.loads(element);
        except Exception as e:
            print(e)
            return None;
    
    @staticmethod
    def json_to_string(element):
        try:
            return json.dumps(element);
        except:
            return None;

    @staticmethod
    def pick_fields(element, array):
        return pydash.pick(element, array);
    @staticmethod
    def omit_fields(element, array):
        return pydash.omit(element, array);

    @staticmethod
    def filter(element, keys, path, condition):
        result = {}
        for key in keys:
            if element[key][path] == condition:
                result[key] = element[key]
        return result;
    @staticmethod
    def pick(element, key, data_type, default_value=None):
            if data_type == str:
                if element is not None and key in element:
                    if bool(str(element[key])) is False:
                        if default_value is not None:
                            return default_value
                    return str(element[key]);
                else:
                    if default_value is not None:
                        return default_value
                    return '';
            if data_type == bool:
                if element is not None and key in element:
                    return bool(element[key]);
                else:
                    if default_value is not None:
                        return default_value
                    return False;
    
            if data_type == list:
                if element is not None and key in element:
                    if element[key] is None or bool(list(element[key])) is False:
                        if default_value is not None:
                            return default_value
                        else:
                            return []
                    return list(element[key]);
                else:
                    if default_value is not None:
                        return default_value
                    return [];
            if data_type == dict:
                if element is not None and key in element:
                    if element[key] is not None and bool(dict(element[key])) is False:
                        if default_value is not None:
                            return default_value
                    if element[key] is not None:
                        return dict(element[key]);
                    return None;
                else:
                    if default_value is not None:
                        return default_value
                    return None;
            if data_type == int:
                if element is not None and key in element:
                    return int(element[key]);
                else:
                    if default_value is not None:
                        return default_value
                    return 0;
            if data_type == round:
                if element is not None and key in element:
                    return round(int(str(element[key])));
                else:
                    if default_value is not None:
                        return default_value
                    return 0;
    
            if data_type == float:
                if element is not None and key in element:
                    r = float(str(element[key]))
                    print(f"{key} {r} {type(r)} {element[key]}")
                    if pydash.is_nan(r) or str(r).lower()=='nan':
                        return 0.0
                    return r;
                else:
                    if default_value is not None:
                        return default_value
                    return 0.0;
            return None;

    @staticmethod
    def filter_by_fields_(data, fields: list[str]) -> dict:
        """Filter dictionary by keeping only specified fields, including nested paths"""
        if not data or not fields:
            return {}
        if isinstance(data, list):
            print('filtering list', data, fields)
            return [DictUtils.filter_by_fields(item, fields) for item in data]
        
        if not isinstance(data, dict):
            print('filtering not dict', data, fields)
            return data;
        result = {}
        print('filtering', data, fields)
        for field in fields:
            if '.' in field:
                # Handle nested fields
                parts = field.split('.')
                value = DictUtils.remove_none_values(data)
                for part in parts:
                    value = value.get(part, {}) if isinstance(value, dict) else {}
                if value:
                    pydash.set_(result, field, value)
            else:
                # Handle top-level fields
                if field in data:
                    result[field] = data[field]
        return DictUtils.remove_none_values(result)
    @staticmethod
    def remove_none_values(obj, keep_empty=True):
        """
        Recursively remove all None values from dictionaries and lists
        
        Args:
            obj: Dict, list, or other object to clean
            keep_empty: Boolean to determine if empty dicts/lists should be kept
            
        Returns:
            Cleaned object with all None values removed
        """
        if isinstance(obj, dict):
            cleaned = {
                key: DictUtils.remove_none_values(value, keep_empty)
                for key, value in obj.items()
                if value is not None and DictUtils.remove_none_values(value, keep_empty) is not None
            }
            return cleaned if cleaned or keep_empty else None
        
        if isinstance(obj, list):
            cleaned = [
                DictUtils.remove_none_values(item, keep_empty)
                for item in obj
                if item is not None and DictUtils.remove_none_values(item, keep_empty) is not None
            ]
            return cleaned if cleaned or keep_empty else None
        
        if isinstance(obj, (str, int, float, bool)):
            return obj
            
        return None
    @staticmethod
    def filter_by_fields_(obj, path):
        """
        Extract a value from a nested object using a dot-separated path.
        
        Args:
            obj: The object to extract from (dict, list, or any nested structure)
            path: String with dot-separated path (e.g., "user.address.street")
        
        Returns:
            The value at the specified path or None if path doesn't exist
        """
        keys = path.split('.')
        result = obj
        
        try:
            for key in keys:
                if isinstance(result, (dict, list)):
                    # Handle array indices
                    if key.isdigit() and isinstance(result, list):
                        result = result[int(key)]
                    else:
                        result = result[key]
                else:
                    return None
            return result
        except (KeyError, IndexError, TypeError):
            return None
    def filter_by_fields(obj: Any, paths: Union[List[str], str]) -> Any:
        # Convert single path to list
        if isinstance(paths, str):
            paths = [paths]
        
        # Convert paths to set for O(1) lookup
        path_set = set(paths)
            
        def should_include_value(value: Any) -> bool:
            """Check if value should be included in output"""
            return isinstance(value, (str, int, float, bool, DatetimeWithNanoseconds)) or value is None
        
        def pick_nested(obj: Any, current_path: str = '') -> Any:
            """Recursively pick values from nested structure"""
            # Handle lists
            if isinstance(obj, list):
                picked_list = [pick_nested(item, current_path) for item in obj]
                return [item for item in picked_list if item not in (None, {}, [])]
                
            # Handle non-dict values
            if not isinstance(obj, dict):
                return obj if should_include_value(obj) else None
                
            # Handle dict
            result = {}
            for key, value in obj.items():
                # if key == 'ages' and key in path_set:
                    # print('ages ==>', value)
                if key in path_set and type(value) is list and len(value) == 0:
                    result[key] = value
                    continue
                new_path = f"{current_path}.{key}" if current_path else key
                
                # Check if this exact path is requested
                # if new_path == 'ages.text':
                    # print('ages.text ==>', value)
                if new_path in path_set:
                    if should_include_value(value):
                        result[key] = value
                        continue
                    
                # Check if this path is a prefix of any requested path
                if any(p.startswith(new_path + '.') for p in path_set):
                    # if new_path == 'ages':
                        # print('ages.text ==>', value)
                    picked = pick_nested(value, new_path)
                    if picked not in (None, {}, []):
                        result[key] = picked
                        
            return result
        
        return pick_nested(obj)

def get_nested_value(data, path):
    keys = path.split('.')
    value = data
    for key in keys:
        if isinstance(value, dict):
            value = value.get(key)
        else:
            return None
    return value

class ArrayUtils:

    @staticmethod
    def compare_lists(list1: list, list2: list, path: str) -> dict:
        """
        Compare two lists and return the changes.

        Args:
            list1 (list): The first list to compare.
            list2 (list): The second list to compare.
            path (str): The current path in the nested structure.

        Returns:
            dict: A dictionary containing the paths of changed fields and their values.
        """
        changes = {}
        max_length = max(len(list1), len(list2))

        for index in range(max_length):
            current_path = f"{path}[{index}]"
            value1 = list1[index] if index < len(list1) else None
            value2 = list2[index] if index < len(list2) else None

            if value1 != value2:
                changes[current_path] = {'old_value': value1, 'new_value': value2}

        return changes
    @staticmethod
    def filter_array_with_another(array_to_filter, filter):
        return pydash.filter_(array_to_filter, lambda x: pydash.find(filter, lambda y: x==y) is not None);

    def filter(array, filter):
        return pydash.filter_(array, filter);

    @staticmethod
    def get_uniq_value(elements, _key, _key_=None, parse=False):
        keys = [];
        for element in elements:
            if parse is False:
                print('type', type(element[_key]))
                if type(element[_key]) == str or type(element[_key]) == dict:
                    keys.append(element[_key]);
                elif type(element[_key]) == list:
                    keys = pydash.arrays.concat(keys, element[_key])
                    print('here', keys)
            else:
                data = json.loads(JsonEncoder().encode(element))[_key]
                if type(element[_key]) == str or type(element[_key]) == dict:
                    keys.append(data)
                elif type(element[_key]) == list:
                    keys = pydash.arrays.concat(keys, data)
        if _key_ is None:
            return pydash.uniq(keys);
        else:
            return pydash.uniq_by(keys, _key_);
    @staticmethod
    def group_elements_by_key(elements, _key, parse=False):
        keys = ArrayUtils.get_uniq_value(elements, _key, parse);
        result = {};
        for key in keys:
            e = None;
            if parse is False:
                e = pydash.filter_(elements, lambda x: x[_key]==key);
            else:
                e = pydash.filter_(elements, lambda x: json.loads(JsonEncoder().encode(x))[_key]==key);
            if key not in result:
                result[key] = pydash.flatten([e]);
            else:
                result[key] = pydash.flatten(result[key].append(e));
        return result;
    
    @staticmethod
    def sum_grouped_by_key(elements, _key):
        result = []
        for key in pydash.keys(elements):
            sum = pydash.sum_by(elements[key], _key);
            result.append({'key': key, "sum": sum})
        return result;

    @staticmethod
    def string_to_array(element) -> 'list':
        try:
            return json.loads(element);
        except:
            return None;
    
    @staticmethod
    def join_string_to_array(element, separator='') -> 'Iterable[str]':
        try:
            split = str(element).split(separator);
            if len(split) == 0:
                if len(element) > 0:
                    return [element];
                return []
            else:
                return split;
        except:
            return [];
    
    @staticmethod
    def array_include_array(element, array_check, x = None, y = None) -> 'bool':
        
        def check_filter(check_data, filter_data):
            if x is None or bool(x) is False:
                if y is None or bool(y) is False:
                    return check_data == filter_data
                else:
                    if y in check_data:
                        return check_data[y] == filter_data
                    return False
            else:
                if y is None or bool(y) is False:
                    if x in filter_data:
                        return check_data == filter_data[x]
                    return False
                else:
                    if x in filter_data and y in check_data:
                        return filter_data[x] == check_data[y]
                    return False
                
        def filter(filter_data):
            return pydash.filter_(array_check, lambda arr: check_filter(arr,filter_data))
                 
        return pydash.filter_(element,  lambda element_data: filter(element_data))

    @staticmethod
    def array_to_string(element) -> 'str':
        try:
            return json.dumps(element);
        except:
            return None;
    @staticmethod
    def pick(element, pickable_fields) -> 'list':
        return pydash.filter_(element, lambda x: pydash.includes(pickable_fields, x))
    
    @staticmethod
    def pick_iterable_childs_with_key(iterable, pickable_fields) -> 'list':
        values = [];
        
        if iterable is None or type(iterable) is not list:
            return [];
    
        for element in iterable:
            picked = pydash.pick(element,  pickable_fields);
            if picked is not None and bool(picked) is True:
                values.append(picked);
        return values
    
    @staticmethod
    def pick_iterable_values_with_key(iterable, field) -> 'list':
        values = [];
        
        if iterable is None or type(iterable) is not list:
            return [];
    
        for element in iterable:
            if field in element and bool(element[field]) is True:
                values.append(element[field]);
        return values;

    @staticmethod
    def class_list_to_list_dict(classes):
        values = [];
        for class_object in classes:
            to_json = getattr(class_object, "to_json", None)
            if callable(to_json):
                values.append(class_object.to_json());
        return values;
    @staticmethod
    def find(array, predicate):
        return pydash.find(array, predicate)


class RequestUtils:
    @staticmethod
    def get_data(request):
        print(request.data.decode())
        data_request = DictUtils.string_to_json(request.data.decode())
        print("data request",data_request)
        return data_request;


from .display import (
    CampaignDisplay, CampaignDisplayFields, CampaignDisplayFieldsProps,
    AdGroupDisplay, AdGroupDisplayFields, AdGroupDisplayFieldsProps,
    filter_campaign_display_request_fields, filter_adgroup_display_request_fields,
    ADGROUP_DISPLAY_PICKABLE_FIELDS, CAMPAIGN_DISPLAY_PICKABLE_FIELDS, CAMPAIGN_DISPLAY_NOT_PICKABLE_FIELDS
    )
from .models import ClientCustomerId
from .models.user_interest import UserInterest
from .utils import search_youtube_channel_by_id, search_youtube_video_by_id




from dataclasses import dataclass
from .....utils.utils import DictUtils
import os

CLIENT_COLLECTION = os.environ.get('CLIENT_COLLECTION')
@dataclass
class ClientFields:
    id = "id"
    name = "name"
    uid = "uid"
    description = "description"
    client_id = "client_id"
    client_secret = "client_secret"
    redirect_uris = "redirect_uris"
    allowed_redirect_uris = "allowed_redirect_uris"
    uri = "uri"
    grant_types = "grant_types"
    response_types = "response_types"
    token_endpoint_auth_method  = "token_endpoint_auth_method"
    token_endpoint_auth_methods = "token_endpoint_auth_methods"
    scopes = "scopes"
    scope = "scope"
    client_id_issued_at = "client_id_issued_at"
    default_redirect_uri = "default_redirect_uri"
    created_at = "createdAt"
    allow_direct_grant_access = "allow_direct_grant_access"
    client_type = "client_type"

    @staticmethod
    def keys():
        return DictUtils.get_keys(ClientFieldProps);

    @staticmethod
    def filtered_keys(field, condition=True):
        mutable = DictUtils.filter(ClientFieldProps, DictUtils.get_keys(ClientFieldProps), field, condition)
        return DictUtils.get_keys(mutable);

ClientFieldProps = {
     ClientFields.id: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    ClientFields.name: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    ClientFields.uid: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    ClientFields.description: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    ClientFields.client_id: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    ClientFields.uri: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    ClientFields.default_redirect_uri: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    ClientFields.scope: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    ClientFields.token_endpoint_auth_method: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    ClientFields.client_secret: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    ClientFields.redirect_uris: {
        "type": list,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": [],
        "pickable": True
    },
    ClientFields.allowed_redirect_uris: {
        "type": list,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": [],
        "pickable": True
    },
    ClientFields.grant_types: {
        "type": list,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": [],
        "pickable": True
    },
    ClientFields.response_types: {
        "type": list,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": [],
        "pickable": True
    },
    ClientFields.scopes: {
        "type": list,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": [],
        "pickable": True
    },
    ClientFields.client_id_issued_at: {
        "type": int,
        "required": True,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "default_value": 0,
        "pickable": True
    },
    ClientFields.created_at: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "default_value": 0,
        "pickable": True
    },
    ClientFields.allow_direct_grant_access: {
        "type": bool,
        "required": True,
        "mutable": True,
        "editable":True,
        "interactive": True,
        "default_value": False,
        "pickable": True
    },
    ClientFields.token_endpoint_auth_methods: {
        "type": list,
        "required": True,
        "mutable": True,
        "editable":True,
        "interactive": True,
        "default_value": [],
        "pickable": True
    },
    ClientFields.client_type: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable":False,
        "interactive": True,
        "default_value": "public",
        "pickable": True
    }
}

STANDARD_FIELDS = ClientFields.filtered_keys('pickable', True)


from dataclasses import dataclass
from typing import Any
from  ....utils.utils import BaseFieldsClass
from google.cloud.firestore_v1.document import Timestamp
from google.api_core.datetime_helpers import DatetimeWithNanoseconds


@dataclass
class BaseCampaignFields(BaseFieldsClass):
    id = "id"
    id_campagne = "id_campagne"
    name = "name"
    objective = "objective"
    adChannel = "adChannel"
    status = "status"
    startDate = "startDate"
    endDate = "endDate"
    startDateFormattedGoogle = "startDateFormattedGoogle"
    endDateFormattedGoogle = "endDateFormattedGoogle"
    areaTargetedOption = "areaTargetedOption"
    areaExcludedOption = "areaExcludedOption"
    budget = "budget"
    budgetId = "budgetId"
    dailyBudget = "dailyBudget"
    bid = "bid"
    urlPromote = "urlPromote"
    strategie = "strategie"
    deliveryMethod = "deliveryMethod"
    trackingTemplate = "trackingTemplate"
    finalUrlSuffix = "finalUrlSuffix"
    urlCustomParameters = "urlCustomParameters"
    clientCustomerId = "clientCustomerId"
    servingStatus = "servingStatus"
    owner = "owner"
    createdAt = "createdAt"
    createdBy = "createdBy"
    endedAt = "endedAt"
    isPayed = "isPayed"
    isEnded = "isEnded"
    type = "type"
    accountId = "accountId"
    is_removed = "is_removed"
    targetedLocations = "targetedLocations"
    excludedLocations = "excludedLocations"
    ages = "ages"
    genders = "genders"
    devicesTargeted = "devicesTargeted"
    devicesExcluded = "devicesExcluded"
    adsSchedules = "adsSchedules"
    budgetEnded = "budgetEnded"
    publish = "publish"
    publishing = "publishing"
    ad_group_id = "ad_group_id"
    ad_group_id_firebase = "ad_group_id_firebase"
    publicationDate = "publicationDate"
    provider = "provider"
    isComplete = "isComplete"

    def __init__(self, **kwargs):
        self.class_fields_props = BaseCampaignFieldsProps
        super().__init__(**kwargs)



BaseCampaignFieldsProps = {
    BaseCampaignFields.id: {
        "internal": False,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseCampaignFields.name: {
        "internal": True,
        "type": str,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseCampaignFields.status: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": "ENABLED"
    },
    BaseCampaignFields.startDate: {
         "internal": True,
        "type": str,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseCampaignFields.endDate: {
         "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseCampaignFields.startDateFormattedGoogle: {
         "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": False,
        "default_value": None
    },
    BaseCampaignFields.endDateFormattedGoogle: {
         "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": False,
        "default_value": None
    },
    BaseCampaignFields.strategie: {
         "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": "CPM"
    },
    BaseCampaignFields.bid: {
        "internal": True,
        "type": float,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": False,
        "default_value": 0.1
    },
    BaseCampaignFields.budget: {
        "internal": True,
        "type": [int,float],
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": 4
    },
    BaseCampaignFields.dailyBudget: {
        "internal": True,
        "type": [int,float],
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": 4
    },
    BaseCampaignFields.budgetId: {
        "internal": True,
        "type": int,
        "required": False,
        "mutable": False,
        "editable": True,
        "interactive": True,
        "pickable": False,
        "default_value": 0
    },
    BaseCampaignFields.urlPromote: {
        "internal": True,
        "type": str,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseCampaignFields.objective: {
        "internal": True,
        "type": dict,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": {"obj_id": "6"}
    },
    BaseCampaignFields.adChannel: {
        "type": dict,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": {"obj_id": 'display'}
    },
    BaseCampaignFields.provider: {
        "internal": True,
        "type": str,
        "required": True,
        "mutable": False,
        "editable": True,
        "interactive": True,
        "pickable": False,
        "default_value": None
    },
    BaseCampaignFields.type: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseCampaignFields.owner: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseCampaignFields.accountId: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseCampaignFields.id_campagne: {
        "internal": True,
        "type": int,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": 0
    },
    BaseCampaignFields.ad_group_id: {
        "internal": True,
        "type": int,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": 0
    },
    BaseCampaignFields.ad_group_id_firebase: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseCampaignFields.isEnded: {
        "internal": True,
        "type": bool,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": False,
        "default_value": False
    },
    BaseCampaignFields.isPayed: {
        "internal": True,
        "type": bool,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": False,
        "default_value": False
    },
    BaseCampaignFields.is_removed: {
        "internal": True,
        "type": bool,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": False,
        "default_value": False
    },
    BaseCampaignFields.finalUrlSuffix: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseCampaignFields.trackingTemplate: {
         "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseCampaignFields.urlCustomParameters: {
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    BaseCampaignFields.publish: {
        "internal": True,
        "type": bool,
        "required": False,
        "mutable": False,
        "editable": True,
        "interactive": True,
        "pickable": False,
        "default_value": False
    },
    BaseCampaignFields.publishing: {
        "internal": True,
        "type": bool,
        "required": False,
        "mutable": False,
        "editable": True,
        "interactive": True,
        "pickable": False,
        "default_value": False
    },
    BaseCampaignFields.isComplete: {
        "internal": True,
        "type": bool,
        "required": False,
        "mutable": False,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": False
    },
    BaseCampaignFields.budgetEnded: {
        "internal": True,
        "type": bool,
        "required": False,
        "mutable": False,
        "editable": True,
        "interactive": True,
        "pickable": False,
        "default_value": False
    },
    BaseCampaignFields.publicationDate: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": False,
        "default_value": ""
    },
    BaseCampaignFields.areaTargetedOption: {
         "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseCampaignFields.areaExcludedOption: {
         "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseCampaignFields.ages: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    BaseCampaignFields.genders: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    BaseCampaignFields.servingStatus: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": True,
        "interactive": True,
        "pickable": False,
        "default_value": None
    },
    BaseCampaignFields.clientCustomerId: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseCampaignFields.createdBy: {
         "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": False,
        "default_value": None
    },
    BaseCampaignFields.createdAt: {
        "type": [int, str, Timestamp, DatetimeWithNanoseconds],
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseCampaignFields.adsSchedules: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    BaseCampaignFields.deliveryMethod: {
         "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": False,
        "default_value": None
    },
    BaseCampaignFields.targetedLocations: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    BaseCampaignFields.excludedLocations: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    BaseCampaignFields.devicesTargeted: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    BaseCampaignFields.devicesExcluded: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    }
}


STANDARD_FIELDS = BaseCampaignFields(fields_props=BaseCampaignFieldsProps).filtered_keys('pickable', True)
BASE_CAMPAIGN_PICKABLE_FIELDS = BaseCampaignFields(fields_props=BaseCampaignFieldsProps).filtered_keys('pickable', True) + ['objective.primary']



from ...base.firebase_collection import FirebaseCollectionBase
from ....utils import (DictUtils, DateUtils, get_object_model_class, init_class_kwargs, BaseClass, get_request_fields, RequestFields)
from ....utils.response import ApiResponse, Error, ResponseStatus, StatusCode
from .base_campaign_fields import BaseCampaignFieldsProps, STANDARD_FIELDS, BASE_CAMPAIGN_PICKABLE_FIELDS
from typing import Any
from dataclasses import dataclass

from .objective import Objective
from .adchannel import AdChannel
from .ages import Ages
from .genders import Genders
from .locations import Location
from .adschedules import AdsSchedules
from .custom_parameters import CustomParameters
from .devices import Devices


def filter_request_fields(fields, default_fields = BASE_CAMPAIGN_PICKABLE_FIELDS):
    request_fields = get_request_fields(fields, default_fields, [default_fields[0]])
    if request_fields is None or bool(request_fields) is False:
        request_fields = [default_fields[0]]
    return request_fields



@dataclass(init=False)
class BaseCampaign(FirebaseCollectionBase):
    id: str
    id_campagne: int
    name: str
    status: str
    objective: Objective
    adChannel: AdChannel
    startDate: str
    endDate: str
    startDateFormattedGoogle: str
    endDateFormattedGoogle: str
    areaTargetedOption: str
    areaExcludedOption: str
    budget: int
    budgetId: int
    dailyBudget: int
    bid: float
    urlPromote: str
    strategie: str
    ages: list[Ages]
    genders: list[Genders]
    targetedLocations: list[Location]
    excludedLocations: list[Location]
    devicesTargeted: list[Devices]
    devicesExcluded: list[Devices]
    deliveryMethod: str
    trackingTemplate: str
    finalUrlSuffix: str
    accountId: str
    type: str
    is_removed: bool
    isEnded: bool
    isComplete: bool
    isPayed: bool
    publish: bool
    publishing: bool
    budgetEnded: bool
    ad_group_id: int
    ad_group_id_firebase: str
    servingStatus: str
    owner: str
    provider: str
    urlCustomParameters: list[CustomParameters]
    adsSchedules: list[AdsSchedules]
    createdAt: any
    createdBy: any
    clientCustomerId: str
    __baseFields: BaseCampaignFieldsProps

    def __init__(self, campaign=None, collection_name=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, campaign, STANDARD_FIELDS, BaseCampaignFieldsProps, collection_name, ['id'], **kwargs)
        super().__init__(**data_args);
        for key in keys:
            if key == "objective":
                setattr(self, key, Objective().from_dict(cls_object[key]))
            elif key == "adChannel":
                setattr(self, key, AdChannel().from_dict(cls_object[key]))
            elif key == "ages":
                setattr(self, key, Ages().fromListDict(cls_object[key]))
            elif key == "genders":
                setattr(self, key, Genders().fromListDict(cls_object[key]))
            elif key == "targetedLocations":
                setattr(self, key, Location().fromListDict(cls_object[key]))
            elif key == "excludedLocations":
                setattr(self, key, Location().fromListDict(cls_object[key]))
            elif key == "devicesTargeted":
                setattr(self, key, Devices().fromListDict(cls_object[key]))
            elif key == "devicesExcluded":
                setattr(self, key, Devices().fromListDict(cls_object[key]))
            elif key == "urlCustomParameters":
                setattr(self, key, CustomParameters().fromListDict(cls_object[key]))
            elif key == "adsSchedules":
                setattr(self, key, AdsSchedules().fromListDict(cls_object[key]))
            elif key == "createdAt":
                # print('created at int', type(cls_object[key]))
                setattr(self, key, cls_object[key]) 
            else:
                setattr(self, key, cls_object[key]) 


    @staticmethod
    def generate_model(_key_="default_value"):
        campaign = {};
        props = BaseCampaignFieldsProps
        for k in DictUtils.get_keys(props):
            campaign[k] = props[k][_key_];
        return campaign;

    
    def to_dict(self, fields=None):
        createdAt = DateUtils.convert_firestore_timestamp_to_str(self.createdAt)
        return DictUtils.remove_none_values({
            "id": self.id,
            "id_campagne": self.id_campagne,
            "name": self.name,
            "status": self.status,
            "objective": self.objective.to_dict(None),
            "adChannel": self.adChannel.to_dict(None),
            "startDate": self.startDate,
            "endDate": self.endDate,
            "startDateFormattedGoogle": self.startDateFormattedGoogle,
            "endDateFormattedGoogle": self.endDateFormattedGoogle,  
            "areaTargetedOption": self.areaTargetedOption,
            "areaExcludedOption": self.areaExcludedOption,  
            "budget": self.budget,
            "budgetId": self.budgetId,
            "dailyBudget": self.dailyBudget,
            "bid": self.bid,
            "urlPromote": self.urlPromote,
            "strategie": self.strategie,
            "ages": Ages().toListDict(self.ages, None),
            "genders": Genders().toListDict(self.genders, None),
            "targetedLocations": Location().toListDict(self.targetedLocations, None),
            "excludedLocations": Location().toListDict(self.excludedLocations, None),
            "devicesTargeted": Devices().toListDict(self.devicesTargeted, None),
            "devicesExcluded": Devices().toListDict(self.devicesExcluded, None),
            "deliveryMethod": self.deliveryMethod,
            "trackingTemplate": self.trackingTemplate,
            "finalUrlSuffix": self.finalUrlSuffix,
            "accountId": self.accountId,
            "type": self.type,
            "is_removed": self.is_removed,
            "isEnded": self.isEnded,
            "isComplete": self.isComplete,
            "isPayed": self.isPayed,
            "publish": self.publish,
            "publishing": self.publishing,
            "budgetEnded": self.budgetEnded,
            "ad_group_id": self.ad_group_id,
            "ad_group_id_firebase": self.ad_group_id_firebase,
            "servingStatus": self.servingStatus,
            "owner": self.owner,
            "provider": self.provider,
            "urlCustomParameters": CustomParameters().toListDict(self.urlCustomParameters, None),
            "adsSchedules": AdsSchedules().toListDict(self.adsSchedules, None),
            "clientCustomerId": self.clientCustomerId,
            "createdAt": createdAt,
            "createdBy": self.createdBy
        })
        


from ...base.firebase_collection import FirebaseCollectionBase, getTimestamp
from ....utils import (DictUtils, get_object_model_class, init_class_kwargs, BaseClass, get_request_fields, RequestFields)
from ....utils.response import ApiResponse, Error, ResponseStatus, StatusCode
from .account_fields import AccountStatus, AccountFields, AccountLinkFields, AccountLinkFieldProps, AccountFieldProps, STANDARD_FIELDS, ACCOUNT_PICKABLE_FIELDS, LINK_STANDARD_FIELDS, ACCOUNT_COLLECTION, ACCOUNT_LINK_COLLECTION
from typing import Any
from dataclasses import dataclass
import pydash
from adafri.v1.user import User
from google.cloud.firestore_v1.document import DocumentReference
from datetime import datetime
from time import time
def filter_account_request_fields(fields, default_fields = ACCOUNT_PICKABLE_FIELDS):
    request_fields = get_request_fields(fields, default_fields, [default_fields[0]])
    if request_fields is None or bool(request_fields) is False:
        request_fields = [default_fields[0]]
    return request_fields

@dataclass(init=False)
class Account(FirebaseCollectionBase):
    id: str
    aacid: str
    account_value: int
    creationDate: int
    creationDateString: str
    name: str
    owner: str
    owner_email: str
    status: str
    totalClics: int
    totalCosts: int
    totalImpressions: int
    usedPackTest: bool
    __baseFields: AccountFieldProps

    def __init__(self,_account=None, **kwargs):
        account = _account
        if type(_account) is str:
            account = {"id": _account}
        (cls_object, keys, data_args) = init_class_kwargs(self, account, STANDARD_FIELDS, AccountFieldProps, ACCOUNT_COLLECTION, ['id'], **kwargs)
        super().__init__(**data_args);
        for key in keys:
            setattr(self, key, cls_object[key]) 


    @staticmethod
    def generate_model(_key_="default_value"):
        user = {};
        props = AccountFieldProps
        for k in DictUtils.get_keys(props):
            user[k] = props[k][_key_];
        return user;

    @staticmethod
    def from_dict(account: Any, db=None, collection_name=None) -> 'Account':
        cls_object, keys = get_object_model_class(account, Account, AccountFieldProps);
        if AccountFields.aacid in cls_object and bool(cls_object[AccountFields.aacid]) is False:
            if AccountFields.id in cls_object and bool(cls_object[AccountFields.id]):
                cls_object[AccountFields.aacid] = cls_object[AccountFields.id]
        _account = Account(cls_object, db=db, collection_name=collection_name)
        return _account
    
    def to_dict(self):
        return {
            "id": self.id,
            "aacid": self.aacid,
            "account_value": self.account_value,
            "creationDate": self.creationDate,
            "creationDateString": self.creationDateString,
            "name": self.name,
            "owner": self.owner,
            "owner_email": self.owner_email,
            "status": self.status,
        }
    
    def get(self, fields=None):
        _id = getattr(self, 'aacid', None);
        if bool(_id) is False:
            if self.id is None:
                return None;
            _id = self.id;
        if bool(id) is None:
            return None;
        request_fields = filter_account_request_fields(fields)
        doc = self.document_reference().get();
        if doc.exists is False:
            return None;
        data = {"id": doc.id, **doc.to_dict()}
        return Account.from_dict(account=Account.from_dict(account=data).to_dict(),  db=self.db, collection_name=self.collection_name);


    def getUserAccountsLinked(self, uid, fields=None):
        accounts_link = []
        request_fields = filter_account_request_fields(fields)
        # print('account query', account)
        links_ids = []
        accounts_links_ = AccountLink(account_link=None).query(AccountLink, "account_link", query_params=[{"key": "target", "comp": "==", "value": uid}])
        for account in accounts_links_:
            links_ids.append(account.linked_account.account.id)
        accounts_batch = self.queryBatch(Account, "account", links_ids)
        for account in accounts_batch:
            accounts_link.append(account)
            # print(account.linked_account.account.id)
        return accounts_link
    def getUserAccounts(self, uid, includeLinks = True, fields=None):
        accounts = []
        accounts_link = []
        accounts_ = self.query(Account, "account", query_params=[{"key": "owner", "comp": "==", "value": uid}])
        request_fields = filter_account_request_fields(fields)
        # print('account query', account)
        for account in accounts_:
            accounts.append(account)
        if includeLinks is True:
            accounts_link.extend(self.getUserAccountsLinked(uid))
            # print(account.linked_account.account.id)
        return accounts, accounts_link
        # accounts_link.append(account.to_json(['id']))
    def getByUserId(self,id):
        accounts = [];
        docs = self.collection().where('owner','==',id).get();
        for doc in docs:
            data = doc.to_dict();
            data['id'] = doc.id;
            data['aacid'] = doc.id;
            print(data)
            accounts.append(Account.from_dict(data))
        return accounts;
    # def query(self, query_params: list, first=False, limit=None):
    #     result = [];
    #     query_result = self.custom_query(query_params, first=first, limit=limit)
    #     if bool(query_result):
    #         if first:
    #             return Account.from_dict(account=query_result, db=self.db, collection_name=self.collection_name)
    #         else:
    #             for doc in query_result:
    #                 result.append(Account.from_dict(account=doc, db=self.db, collection_name=self.collection_name))
    #             return result
    #     if first:
    #             return None
    #     return [];

    def getAccount(self):
            id = self.id
            if bool(id) is False:
                if bool(self.aacid) is False:
                    return None;
                id = self.aacid
            account = self.get(self.id);
            if account is None:
                return None;
            return account.to_dict();
    def create(self, **kwargs):
        data_dict = DictUtils.pick_fields(kwargs, AccountFields(fields_props=AccountFieldProps).filtered_keys('mutable', True));
        account_model = Account().from_dict(DictUtils.merge_dict(data_dict, Account.generate_model()));
        if bool(account_model.to_json()) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Empty request","INVALID_REQUEST", 1))
        
        user = None
        if 'uid' in kwargs:
            user = User(kwargs['uid']).get();
        if user is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("User not found","INVALID_REQUEST", 1));
    
        account_model.owner = user.id;
        account_model.owner_email = user.email;
        name = account_model.name

        if bool(name) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Password required","INVALID_REQUEST", 1));
        docs = Account().query(self, "account", query_params=[{"key": "owner", "comp": "==", "value": account_model.owner}, {"key": "name", "comp": "==", "value": name}]);
        found_users = [];
        if len(docs) > 0:
            for doc in docs:
                found_users.append(Account.from_dict(doc.to_dict()).to_json());
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Account with name {account_model.email} already exist","INVALID_REQUEST", 1));
    
    
        try:
            doc: DocumentReference = Account().collection().document()
            account_model.aacid = doc.id;
            account_model.id = doc.id;
            account_model.status = AccountStatus.ACTIVE;
            creationDate = datetime.now()
            creationDateN = int(creationDate.timestamp()) * 1000
            creationDateString = creationDate.isoformat();
            creationDateString = creationDate.strftime("%d %B %Y à %H:%M:%S UTC");
            account_model.creationDate = creationDateN;
            account_model.creationDateString = creationDateString;
            doc.set({**account_model.to_json(), "createdAt": getTimestamp()}, merge=True)
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, DictUtils.filter_by_fields(account_model.to_json(), ['id', 'aacid', 'name']), None);
        except Exception as e:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(str(e),"INVALID_REQUEST", 1));

    def update(self, data):
        try:
            last_value = self.to_json();
            filtered_value = pydash.pick(data, AccountFields.filtered_keys('editable', True));
            new_value = DictUtils.merge_dict(filtered_value, self.to_json());
            changed_fields = DictUtils.get_changed_field(last_value, new_value);
            data_update = DictUtils.dict_from_keys(filtered_value, changed_fields);
            if bool(data_update) is False:
                return None;
            if 'account_value' in data_update:
                data_update['account_value'] = float(str(data_update['account_value']))
            self.document_reference().set(data_update, merge=True)
            return DictUtils.dict_from_keys(self.getAccount(), changed_fields);
        except Exception as e:
            print(e)
            return None;

    def update_account_value(self, params):
        if 'value' not in params:
            return None;
        if 'operation' not in params:
            return None;
        account_value = params['value'];
        operation = params['operation'];
    
        try:
            value = float(str(account_value))
            current_value = float(str(DictUtils.pick(self.to_json(), "account_value", float, 0)))
            new_value = None;
            if str(operation).lower() == 'add':
                new_value = value + current_value;
            elif str(operation).lower() == 'remove':
                new_value = current_value - value;
            if new_value is None or new_value < 0:
                return None;
        
            data_update = {'account_value': new_value, 'last_value': current_value}
            print('new', data_update)
            self.document_reference(self.id).set(data_update, merge=True)
            return DictUtils.dict_from_keys(self.getAccount(), ['account_value']);
        except Exception as e:
            print('exception', e)
            return None;

    def remove(self, only_mark_as_removed=True):
        try:
            if self.id is None:
                return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Cannot identify Account with id {self.id}","INVALID_REQUEST", 1));
            if only_mark_as_removed:
                self.document_reference().set({"is_removed": True})
            else:
                self.document_reference().delete();
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, {"message": f"Account {self.id} deleted"}, None);
        except:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"An error occurated while removing authorization code with id {self.id}","INVALID_REQUEST", 1));


@dataclass(init=False)
class UserRole(BaseClass):
    admin: bool
    readOnly: bool
    def __init__(self, role=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, role, ['admin', "readOnly"], None, None, [], **kwargs)
        for key in keys:
            setattr(self, key, cls_object[key]) 
    def to_dict(self, field=None):
        return {
            "admin": self.admin,
            "readOnly": self.readOnly
        }
    @staticmethod
    def from_dict(obj: Any) -> 'UserRole':
        cls_object, keys = get_object_model_class(obj, UserRole, None);
        return UserRole(cls_object)
@dataclass(init=False)
class LinkedAccount(BaseClass):
    account: Account
    role: UserRole
    def __init__(self, account_linked=None, **kwargs):
        
        (cls_object, keys, data_args) = init_class_kwargs(self, account_linked, ["account", "role"], None, None, [], **kwargs)
        for key in keys:
            setattr(self, key, cls_object[key])
            if key == 'account':
                # print('cls_object[key]', Account.from_dict(cls_object[key]))
                setattr(self, key, Account.from_dict(cls_object[key]))
            else:
                setattr(self, key, UserRole.from_dict(cls_object[key]))
    def to_dict(self):
        return {
            "account": self.account.to_json(["id"]),
            "role": self.role.to_json()
        }

    @staticmethod
    def from_dict(obj: Any) -> 'LinkedAccount':
        cls_object, keys = get_object_model_class(obj, LinkedAccount, None);
        _account = DictUtils.pick(obj, "account", bool)
        _role = DictUtils.pick(obj, "role", bool)
        return LinkedAccount(cls_object)



@dataclass(init=False)
class AccountLink(FirebaseCollectionBase):
    id: str
    linkDate: int
    linkDateString: str
    owner: str
    owner_email: str
    target: str
    target_email: str
    status: str
    linked_account: LinkedAccount

    def __init__(self, account_link=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, account_link, LINK_STANDARD_FIELDS, AccountLinkFieldProps, ACCOUNT_LINK_COLLECTION, ['id'], **kwargs)
        super().__init__(**data_args);
        for key in keys:
            # print('key', key)
            if key=='linked_account':
                _linked =  DictUtils.pick(cls_object, "linked_account", dict)
                setattr(self, key, LinkedAccount.from_dict(_linked))
            else:
                setattr(self, key, cls_object[key])

    def to_dict(self):
        return self.to_json(['id', "linked_account"])


    @staticmethod
    def generate_model(_key_="default_value"):
        link = {};
        props = AccountLinkFieldProps
        for k in DictUtils.get_keys(props):
            link[k] = props[k][_key_];
        return link;

    @staticmethod
    def from_dict(account_link: Any, db=None, collection_name=None) -> 'AccountLink':
        cls_object, keys = get_object_model_class(account_link, AccountLink, AccountLinkFieldProps);
        _account = AccountLink(cls_object, db=db, collection_name=collection_name)
        return _account
    
    
    def get(self):
        id = self.id;
        if bool(id) is None:
            return None;
    
        doc = self.document_reference().get();
        if doc.exists is False:
            return None;
        data = {"id": doc.id, **doc.to_dict()()}
        return AccountLink.from_dict(data, db=self.db, collection_name=self.collection_name);

    # def query(self, query_params: list, first=False, limit=None):
    #     result = [];
    #     query_result = self.custom_query(query_params, first=first, limit=limit)
    #     if bool(query_result):
    #         if first:
    #             return AccountLink.from_dict(account=query_result, db=self.db, collection_name=self.collection_name)
    #         else:
    #             for doc in query_result:
    #                 result.append(AccountLink.from_dict(account=doc, db=self.db, collection_name=self.collection_name))
    #             return result
    #     if first:
    #             return None
    #     return [];

    def getAccount(self):
            account = self.get(self.id);
            if account is None:
                return None;
            return account.to_json();

    def update(self, data):
        
        try:
            last_value = self.to_json();
            filtered_value = pydash.pick(data, AccountFields.filtered_keys('editable', True));
            new_value = DictUtils.merge_dict(filtered_value, self.to_json());
            changed_fields = DictUtils.get_changed_field(last_value, new_value);
            data_update = DictUtils.dict_from_keys(filtered_value, changed_fields);
            if bool(data_update) is False:
                return None;
            self.document_reference().set(data_update, merge=True)
            return DictUtils.dict_from_keys(self.get().to_json(), changed_fields);
        except Exception as e:
            print(e)
            return None;

    def remove(self, only_mark_as_removed=True):
        try:
            if self.id is None:
                return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Cannot identify Account with id {self.id}","INVALID_REQUEST", 1));
            if only_mark_as_removed:
                self.document_reference().set({"is_removed": True})
            else:
                self.document_reference().delete();
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, {"message": f"Account {self.id} deleted"}, None);
        except:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"An error occurated while removing authorization code with id {self.id}","INVALID_REQUEST", 1));


import firebase_admin
from firebase_admin import credentials
from firebase_admin import firestore as f
from firebase_admin.firestore import firestore
from firebase_admin.exceptions import FirebaseError
from firebase_admin.auth import TokenSignError
from firebase_admin.auth import ActionCodeSettings

import os

import pydash

from firebase_admin import auth

def create_firebase_user(email, password) -> 'auth.UserRecord':
    try:
        return auth.create_user(email=email, password=password)
    except ValueError as e:
        raise e
    except FirebaseError as e:
        raise e;  

def create_firebase_custom_token(uid) -> 'auth.UserRecord':
    try:
        return auth.create_custom_token(uid)
    except ValueError as e:
        raise e
    except TokenSignError as e:
        raise e;  

def generate_email_verification_link(email, url) -> 'str':
    try:
        return auth.generate_email_verification_link(email, action_code_settings=ActionCodeSettings(url=url))
    except ValueError as e:
        return None
    except FirebaseError as e:
        return None;  

def generate_password_reset_link(email, url) -> 'str':
    try:
        return auth.generate_password_reset_link(email, action_code_settings=ActionCodeSettings(url=url))
    except ValueError as e:
        return None
    except FirebaseError as e:
        return None;  

def get_user_by_uid(uid) -> 'auth.UserRecord':
    try:
        return auth.get_user(uid)
    except ValueError as e:
        return None;
    except FirebaseError as e:
        return None; 


def get_user_by_email(email) -> 'auth.UserRecord':
    try:
        return auth.get_user_by_email(email)
    except ValueError as e:
        return None;
    except FirebaseError as e:
        return None; 

def get_user_by_phone_number(phone_number) -> 'auth.UserRecord':
    try:
        return auth.get_user_by_phone_number(phone_number)
    except ValueError as e:
        return None;
    except FirebaseError as e:
        return None; 

def update_firebase_user(uid, **kwargs) -> 'auth.UserRecord':
    """
    Keyword Args:
    display_name: The user's display name (optional). Can be removed by explicitly passing
        auth.DELETE_ATTRIBUTE.
    email: The user's primary email (optional).
    email_verified: A boolean indicating whether or not the user's primary email is
        verified (optional).
    phone_number: The user's primary phone number (optional). Can be removed by explicitly
        passing auth.DELETE_ATTRIBUTE.
      photo_url: The user's photo URL (optional). Can be removed by explicitly passing
        auth.DELETE_ATTRIBUTE.
    password: The user's raw, unhashed password. (optional).
    disabled: A boolean indicating whether or not the user account is disabled (optional).
    custom_claims: A dictionary or a JSON string containing the custom claims to be set on the
        user account (optional). To remove all custom claims, pass auth.DELETE_ATTRIBUTE.
    valid_since: An integer signifying the seconds since the epoch (optional). This field is
        set by revoke_refresh_tokens and it is discouraged to set this field directly.
    app: An App instance (optional).

    Returns:
        UserRecord: An updated user record instance for the user."""
    try:
        return auth.update_user(uid, **kwargs)
    except ValueError as e:
        return None
    except FirebaseError as e:
        return None;  

def set_custom_claims(uid, claims):
    try:
        user = auth.get_user(uid)
        current_claims = user.custom_claims;
        if current_claims is None:
            current_claims = {}
        for key in pydash.keys(claims):
            current_claims[key] = claims[key]
        auth.set_custom_user_claims(uid, current_claims)
        docRef = FirestoreApp().firestore_client().collection('users').document(uid)
        docRef.set({"current_claims": current_claims}, merge=True)
        return docRef.get().to_dict()
    except Exception as e:
        return None

DEFAULT_PATH = os.environ.get('FIREBASE_CREDENTIALS_PATH');

def initializeApp(path: str):
    if path is None:
        return None;
    try:
        absolute_path = os.path.dirname(__file__)
        relative_path = path
        # print('relative_path',relative_path)
        if relative_path.startswith('http'):
            credentials_path = relative_path
        else:
            credentials_path = os.path.join(absolute_path, relative_path)
        cred = credentials.Certificate(credentials_path);
        if len(firebase_admin._apps)==0:
            app = firebase_admin.initialize_app(cred);
            return app;
        return firebase_admin.get_app()
    except Exception as e:
        print(e)
        return None
    
class FirestoreApp:
    def __init__(self, credentials_path: str=DEFAULT_PATH) -> None:
        # print("Initializing Firestore from", credentials_path)
        app = initializeApp(credentials_path);
        if app is not None:
            self.app = app;
            
    
    def firestore_client(self) -> firestore.Client:
        app = getattr(self, 'app', None)
        if app is None:
            return None;
        firestore_database = f.client(self.app);
        return firestore_database;



ads_channel_data = [
    {
        "obj_id": 'search',
        "primary": {
        "icon": 'https://storage.googleapis.com/images.adafri.com/gif/rsa.gif',
        "title": 'Annonces sur le moteur de recherche',
        "description":
            'Vos annonces apparaissent dans le moteur de recherche de Google lorsque des clients potentiels recherchent \
            des mots clés liés à vote activité. Vous ne payez que si une personne interagit avec votre annonce.'
        },
        "secondary": {
        "icon": 'pageview',
        "title": 'Annonces sur le moteur de recherche',
        "description":
            'Vos annonces apparaissent dans le moteur de recherche de Google lorsque des clients potentiels recherchent\
            des mots clés liés à vote activité. Vous ne payez que si une personne interagit avec votre annonce.'
        }
    },
    {
        "obj_id": 'display',
        "primary": {
        "icon": 'https://storage.googleapis.com/images.adafri.com/gif/da.gif',
        "title": 'Annonces avec des visuels',
        "description":
            'Travaillez vos annonces avec votre infographe et importez les simplement pour les remettre à la régie de Google.'
        },
        "secondary": {
        "icon": 'web',
        "title": 'Annonces avec des visuels',
        "description":
            'Travaillez vos annonces avec votre infographe et importez les simplement pour les remettre à la régie de Google.'
        }
    },
    {
        "obj_id": 'native',
        "primary": {
        "icon": 'https://storage.googleapis.com/images.adafri.com/gif/rsda.gif',
        "title": 'Pas d\'infographies ? Optez pour le responsive display',
        "description":
            'Une importation simple de votre logo et d\'images et une adaptation automatique aux formats de Google \
            nécessaires à votre diffusion. Il est possible également de rechercher des images à partir d\'un site web;'
        },
        "secondary": {
        "icon": 'web',
        "title": 'Pas d\'infographies ? Optez pour le responsive display',
        "description":
            'Une importation simple de votre logo et d\'images et une adaptation automatique aux formats de Google \
            nécessaires à votre diffusion. Il est possible également de rechercher des images à partir d\'un site web;'
        }
    }
]


from ....base.firebase_collection import FirebaseCollectionBase
from .....utils import (DictUtils, DateUtils, ArrayUtils, get_object_model_class, init_class_kwargs, BaseClass, get_request_fields, RequestFields)
from .....utils.response import ApiResponse, Error, ResponseStatus, StatusCode
from ...data.demographics import ages_data, genders_data
from typing import Any
from dataclasses import dataclass
import pydash
from ...models.base import BaseCampaign, BaseCampaignFieldsProps
from ...models.placement import (Website, YoutubeChannel, YoutubeVideo)
from .display_fields import (CampaignDisplayFields, CampaignDisplayFieldsProps, CAMPAIGN_DISPLAY_PICKABLE_FIELDS, CAMPAIGN_DISPLAY_NOT_PICKABLE_FIELDS, DISPLAY_COLLECTION)
from ...models.display_ads import (DisplayAdsToPublish, ResponsiveDisplayAdsToPublish)
from ...models.placement import Website, YoutubeChannel, YoutubeVideo
from ...models.user_interest import UserInterest
from ...models.client_customer_id import ClientCustomerId
from firebase_admin.firestore import firestore
import time
import random
import string
from datetime import datetime
from ....account import Account
def filter_campaign_display_request_fields(fields, default_fields = CAMPAIGN_DISPLAY_PICKABLE_FIELDS):
    request_fields = get_request_fields(fields, default_fields, [default_fields[0]])
    if request_fields is None or bool(request_fields) is False:
        request_fields = [default_fields[0]]
    return request_fields

@dataclass(init=False)
class CampaignDisplay(BaseCampaign):
    id: str
    targetedPlacements: list[Website]
    excludedPlacements: list[Website]
    websites: list[Website]
    youtubeChannels: list[YoutubeChannel]
    youtubeVideos: list[YoutubeVideo]
    images: list[DisplayAdsToPublish]
    imagesNative: list[ResponsiveDisplayAdsToPublish]
    user_interest: list[UserInterest]
    __baseFields: CampaignDisplayFieldsProps

    def __init__(self, campaign=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, campaign, CAMPAIGN_DISPLAY_PICKABLE_FIELDS, CampaignDisplayFieldsProps, DISPLAY_COLLECTION, ['id'], **kwargs)
        super().__init__(**{**data_args, "campaign": campaign});
        for key in keys:
            if BaseCampaignFieldsProps.get(key) is None:
                website = ['targetedPlacements', 'excludedPlacements', 'websites']
                if key in website:
                    setattr(self, key, Website().fromListDict(cls_object[key]))
                elif key == 'youtubeChannels':
                    setattr(self, key, YoutubeChannel().fromListDict(cls_object[key]))
                elif key == 'youtubeVideos':    
                    setattr(self, key, YoutubeVideo().fromListDict(cls_object[key]))
                elif key == 'images':
                    setattr(self, key, DisplayAdsToPublish().fromListDict(cls_object[key]))
                elif key == 'imagesNative':
                    setattr(self, key, ResponsiveDisplayAdsToPublish().fromListDict(cls_object[key]))
                elif key == 'user_interest':
                    setattr(self, key, UserInterest().fromListDict(cls_object[key]))
                else:
                    setattr(self, key, cls_object[key]) 


    @staticmethod
    def generate_model(_key_="default_value"):
        model = {};
        props = CampaignDisplayFieldsProps
        for k in DictUtils.get_keys(props):
            if k=='ages':
                print('ages', props[k][_key_])
                model[k] = ages_data;
                continue
            if k=='genders':
                print('genders', props[k][_key_])
                model[k] = genders_data;
                continue
            if _key_ in props[k] and props[k][_key_] is not None:
                model[k] = props[k][_key_];
        
        # Generate random campaign name
        random_suffix = ''.join(random.choices(string.ascii_uppercase + string.digits, k=6))
        model['name'] = f'Campaign Display {random_suffix}'
        model['startDate'] = datetime.now().strftime('%Y-%m-%d')
        return model;

    @staticmethod
    def from_dict(campaign: Any, db=None, collection_name=None) -> 'CampaignDisplay':
        cls_object, keys = get_object_model_class(campaign, CampaignDisplay, CampaignDisplayFieldsProps);
        _campaign = CampaignDisplay(cls_object, db=db, collection_name=collection_name)
        return _campaign
    
    def generateCampaign(self, aacid: str, _customerId: str | None = None):
        account = Account(aacid).get()
        if account is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Account not found","INVALID_REQUEST", 1))
        
        customer = None;
        if _customerId is None:
            default_customers = ClientCustomerId().getDefault()
            if(len(default_customers) > 0):
                customer = default_customers[0]
        else:
            customer = ClientCustomerId().get(_customerId)
        if customer is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Customer not found","INVALID_REQUEST", 1))

        # token = guards(['profile'])
        # if token is None:
        #     return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Account not found","PERMISSION_DENIED", 1)).to_json(), StatusCode.status_400
        model = CampaignDisplay.generate_model()
        model['accountId'] = aacid
        model['clientCustomerId'] = customer.customerId
        create = CampaignDisplay().create(model, account, account.owner)
        if create.status == 'error':
            return create
        # print('fields', CAMPAIGN_DISPLAY_PICKABLE_FIELDS)
        values = create.data
        values = DictUtils.omit_fields(values, CAMPAIGN_DISPLAY_NOT_PICKABLE_FIELDS + ['owner', 'id_campagne', 'ad_group_id'])
        # print('values', values)
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, values, None)
    def to_dict(self, fields=None):
        base = super().to_dict(fields)
        return DictUtils.remove_none_values({
            **base,
            "targetedPlacements": Website().toListDict(self.targetedPlacements, None),
            "excludedPlacements": Website().toListDict(self.excludedPlacements, None),
            "websites": Website().toListDict(self.websites, None),
            "youtubeChannels": YoutubeChannel().toListDict(self.youtubeChannels, None),
            "youtubeVideos": YoutubeVideo().toListDict(self.youtubeVideos, None),
            "images": DisplayAdsToPublish().toListDict(self.images, None),
            "imagesNative": ResponsiveDisplayAdsToPublish().toListDict(self.imagesNative, None),
            "user_interest": UserInterest().toListDict(self.user_interest, None),
    })
    def getCampaigns(self, aacid, fields=None):
        campaigns = []
        campaigns_ = self.query(CampaignDisplay, "campaign", query_params=[{"key": "accountId", "comp": "==", "value": aacid}])
        request_fields = filter_campaign_display_request_fields(fields)
        # print('account query', account)
        for campaign in campaigns_:
            campaigns.append(campaign)
            # campaigns.append(D(campaign.to_dict(request_fields)))

        return campaigns
    
    def get(self, fields=None):
        if bool(self.id) is None:
            return None;
        request_fields = filter_campaign_display_request_fields(fields)
        doc = self.document_reference(self.id).get();
        if doc.exists is False:
            return None;
        data = {"id": doc.id, **doc.to_dict()}
        # print('created at===>', DateUtils.convert_firestore_timestamp_to_str(data['createdAt']))
        return CampaignDisplay.from_dict(campaign=data, db=self.db, collection_name=self.collection_name);

    def getByUserId(self,id):
        campaigns = [];
        docs = self.collection().where('owner','==',id).get();
        for doc in docs:
            data = doc.to_dict();
            data['id'] = doc.id;
            campaigns.append(CampaignDisplay.from_dict(data))
        return campaigns;


    def create(self, data, account: any, uid):
        # creation_date = created_at * 1000
        # return DateUtils.from_timestamp(creation_date).isoformat()
        # try:
        data_create = DictUtils.dict_from_keys(data, CampaignDisplayFields().filtered_keys('mutable', True));
        if bool(data_create) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Invalid request","INVALID_REQUEST", 1))
        campaign_dates = DateUtils.format_campaign_date(data_create)
        if "error" in campaign_dates and bool(campaign_dates["error"]) is True:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(campaign_dates['error'],"INVALID_REQUEST", 1))
        data_create = DictUtils.merge_dict({
            **campaign_dates['campaign'],
                "accountId": account.id,
                "owner": account.owner,
                "createdBy": f"users/${uid}"
            }, data_create,True)
        document_id = self.collection().document().id

        exist = self.query(CampaignDisplay, "campaign", [{"key": "name", "comp": "==", "value": data_create['name']}], True)
        if exist is not None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Campaign name {data_create['name']} already exists","INVALID_REQUEST", 1))
        model = CampaignDisplay.generate_model()
        data_create = DictUtils.merge_dict(data_create, model, True)
        doc_ref = self.collection().document(document_id);
        doc_ref.set({**data_create, 'createdAt': firestore.SERVER_TIMESTAMP}, True);
        time.sleep(2) 
        campaign = CampaignDisplay.from_dict({"id": document_id}).get()
        # d = {**data_create, "id": document_id}
        # campaign = CampaignDisplay.from_dict(d).to_dict()
        if campaign is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Failed to create campaign","INVALID_REQUEST", 1))
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, campaign.to_dict(), None);  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the
        # except Exception as e:
        #     print(e)
        #     return {"error": "Exception creating campaign"};
    def update(self, data):
        # try:
        last_value = self.to_dict();
        # print('last', last_value);
        filtered_value = pydash.pick(data, CampaignDisplayFields().filtered_keys('editable', True));
        # print('filtered', filtered_value);
        new_value = DictUtils.merge_dict(filtered_value, last_value, True);
        _changed_fields = DictUtils.compare_nested_objects(last_value, new_value);
        changed_fields = ArrayUtils.filter(DictUtils.get_keys(_changed_fields), lambda x: str(x).find('.') == -1)
        print('changed fields =>', changed_fields)
        data_update = DictUtils.dict_from_keys(filtered_value, changed_fields);
        if bool(data_update) is False:
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, data_update, None);
        if 'startDate' in data_update or 'endDate' in data_update:
            dates = DateUtils.format_campaign_date(new_value, False)
            if dates is None  or ("error" in dates and bool(dates['error']) is True):
                return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(dates['error'],"INVALID_REQUEST", 1))
        
            data_update = {**data_update, **dates['campaign']}
            changed_fields.extend(DictUtils.get_keys(data_update))
        # try:
        #     self.document_reference().set(data_update, merge=True)
        # except Exception as e:
        #     return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(str(e),"INVALID_REQUEST", 1))
        # return DictUtils.dict_from_keys(self.get(), changed_fields);
        fields = changed_fields
        print('fields', fields)
        print('data_update', data_update)
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, data_update, None);
        # return {"status": "ok", "data": CampaignDisplay.from_dict(campaign=data_update).to_dict()};
        # return {"status": "ok", "data": DictUtils.filter_by_fields(self.get().to_dict(), fields)};
        # except Exception as e:
        #     print(e)
        #     return None;

    def remove(self, only_mark_as_removed=True):
        try:
            if self.id is None:
                return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Cannot identify Campaign with id {self.id}","INVALID_REQUEST", 1));
            if only_mark_as_removed:
                self.document_reference().set({"is_removed": True})
            else:
                self.document_reference().delete();
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, {"message": f"Campaign {self.id} deleted"}, None);
        except:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"An error occurated while removing authorization code with id {self.id}","INVALID_REQUEST", 1));




from dataclasses import dataclass
from typing import Any
from  ....utils.utils import DictUtils, ArrayUtils, BaseFieldsClass, BaseClass,init_class_kwargs,get_object_model_class
from ..data.objectives import objectives_data
@dataclass
class ObjectveFields(BaseFieldsClass):
    id = "obj_id"
    primary = "primary"
    secondary = "secondary"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

@dataclass(init=False)
class Objective(BaseClass):
    obj_id: str
    primary: dict
    secondary: dict
    def __init__(self, objective=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, objective, ['obj_id'], None, None, [], **kwargs)
        objective_value = ArrayUtils.find(objectives_data, lambda x: x['obj_id'] == cls_object['obj_id'])
        if objective_value is not None:
            for key in objective_value:
                setattr(self, key, objective_value[key]) 
    def to_dict(self, fields=None):
        return DictUtils.remove_none_values(
            {
            ObjectveFields.id: self.obj_id,
            ObjectveFields.primary: self.primary,
            ObjectveFields.secondary: self.secondary
        }
        )
    @staticmethod
    def from_dict(obj: Any) -> 'Objective':
        cls_object, keys = get_object_model_class(obj, Objective, None);
        return Objective(cls_object)